import copy

import numpy as np
import pandas as pd
import pytest

from napari._tests.utils import check_layer_world_data_extent
from napari.components.dims import Dims
from napari.layers import Surface
from napari.layers.surface.normals import SurfaceNormals
from napari.layers.surface.wireframe import SurfaceWireframe
from napari.utils._test_utils import (
    validate_all_params_in_docstring,
    validate_kwargs_sorted,
)


def test_random_surface():
    """Test instantiating Surface layer with random 2D data."""
    np.random.seed(0)
    vertices = np.random.random((10, 2))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.ndim == 2
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert np.array_equal(layer.vertices, vertices)
    assert np.array_equal(layer.faces, faces)
    assert np.array_equal(layer.vertex_values, values)
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1


def test_random_surface_features():
    """Test instantiating surface layer with features."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    features = pd.DataFrame({'feature': np.random.random(10)})

    data = (vertices, faces, values)
    layer = Surface(data, features=features)
    assert 'feature' in layer.features.columns


def test_set_features_and_defaults():
    """Test setting features and defaults."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)

    data = (vertices, faces, values)
    layer = Surface(data)

    assert layer.features.shape[1] == layer.feature_defaults.shape[1] == 0

    features = pd.DataFrame(
        {
            'str': ('a', 'b') * 5,
            'float': np.random.random(10),
        }
    )
    feature_defaults = pd.DataFrame(
        {
            'str': ('b',),
            'float': (0.5,),
        }
    )

    layer.features = features
    layer.feature_defaults = feature_defaults

    pd.testing.assert_frame_equal(layer.features, features)
    pd.testing.assert_frame_equal(layer.feature_defaults, feature_defaults)


def test_random_surface_no_values():
    """Test instantiating Surface layer with random 2D data but no vertex values."""
    np.random.seed(0)
    vertices = np.random.random((10, 2))
    faces = np.random.randint(10, size=(6, 3))
    data = (vertices, faces)
    layer = Surface(data)
    assert layer.ndim == 2
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert np.array_equal(layer.vertices, vertices)
    assert np.array_equal(layer.faces, faces)
    assert np.array_equal(layer.vertex_values, np.ones(len(vertices)))
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1


def test_random_surface_clearing_vertex_values():
    """Test setting `vertex_values=None` resets values to uniform ones."""
    np.random.seed(0)
    vertices = np.random.random((10, 2))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)
    assert np.array_equal(layer.vertex_values, values)
    layer.vertex_values = None
    assert np.array_equal(layer.vertex_values, np.ones(len(vertices)))


def test_random_3D_surface():
    """Test instantiating Surface layer with random 3D data."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.ndim == 3
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1

    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    assert layer._data_view.shape[1] == 3
    assert layer._view_vertex_values.ndim == 1


def test_random_4D_surface():
    """Test instantiating Surface layer with random 4D data."""
    np.random.seed(0)
    vertices = np.random.random((10, 4))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.ndim == 4
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1

    layer._slice_dims(Dims(ndim=4, ndisplay=3))
    assert layer._data_view.shape[1] == 3
    assert layer._view_vertex_values.ndim == 1


def test_random_3D_timeseries_surface():
    """Test instantiating Surface layer with random 3D timeseries data."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random((22, 10))
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.ndim == 4
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1
    assert layer.extent.data[1][0] == 21

    layer._slice_dims(Dims(ndim=4, ndisplay=3))
    assert layer._data_view.shape[1] == 3
    assert layer._view_vertex_values.ndim == 1

    # If a values axis is made to be a displayed axis then no data should be
    # shown
    with pytest.warns(UserWarning, match='Assigning multiple data per vertex'):
        layer._slice_dims(Dims(ndim=4, ndisplay=3, order=(3, 0, 1, 2)))
    assert len(layer._data_view) == 0


def test_random_3D_multitimeseries_surface():
    """Test instantiating Surface layer with random 3D multitimeseries data."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random((16, 22, 10))
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.ndim == 5
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1
    assert layer.extent.data[1][0] == 15
    assert layer.extent.data[1][1] == 21

    layer._slice_dims(Dims(ndim=5, ndisplay=3))
    assert layer._data_view.shape[1] == 3
    assert layer._view_vertex_values.ndim == 1


def test_changing_surface():
    """Test changing surface layer data"""
    np.random.seed(0)
    vertices = np.random.random((10, 2))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)

    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer.data = data
    assert layer.ndim == 3
    assert np.all(
        [
            np.array_equal(ld, d)
            for ld, d in zip(layer.data, data, strict=False)
        ]
    )
    assert layer._data_view.shape[1] == 2
    assert layer._view_vertex_values.ndim == 1

    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    assert layer._data_view.shape[1] == 3
    assert layer._view_vertex_values.ndim == 1


def test_visiblity():
    """Test setting layer visibility."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.visible is True

    layer.visible = False
    assert layer.visible is False

    layer = Surface(data, visible=False)
    assert layer.visible is False

    layer.visible = True
    assert layer.visible is True


def test_surface_gamma():
    """Test setting gamma."""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)
    assert layer.gamma == 1

    # Change gamma property
    gamma = 0.7
    layer.gamma = gamma
    assert layer.gamma == gamma

    # Set gamma as keyword argument
    layer = Surface(data, gamma=gamma)
    assert layer.gamma == gamma


def test_world_data_extent():
    """Test extent after applying transforms."""
    data = [(-5, 0), (0, 15), (30, 12)]
    min_val = (-5, 0)
    max_val = (30, 15)
    layer = Surface((np.array(data), np.array((0, 1, 2)), np.array((0, 0, 0))))
    extent = np.array((min_val, max_val))
    check_layer_world_data_extent(layer, extent, (3, 1), (20, 5))


def test_shading():
    """Test setting shading"""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)
    layer = Surface(data)

    # change shading property
    shading = 'flat'
    layer.shading = shading
    assert layer.shading == shading

    # set shading as keyword argument
    layer = Surface(data, shading=shading)
    assert layer.shading == shading


def test_texture():
    """Test setting texture"""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)

    texture = np.random.random((32, 32, 3)).astype(np.float32)
    texcoords = vertices[:, :2]
    layer = Surface(data, texture=texture, texcoords=texcoords)

    np.testing.assert_allclose(layer.texture, texture)
    np.testing.assert_allclose(layer.texcoords, texcoords)
    assert layer._has_texture

    layer.texture, layer.texcoords = None, texcoords
    assert not layer._has_texture

    layer.texture, layer.texcoords = texture, None
    assert not layer._has_texture

    layer.texture, layer.texcoords = None, None
    assert not layer._has_texture

    layer.texture, layer.texcoords = texture, texcoords
    assert layer._has_texture


def test_vertex_colors():
    """Test setting vertex colors"""
    np.random.seed(0)
    vertices = np.random.random((10, 3))
    faces = np.random.randint(10, size=(6, 3))
    values = np.random.random(10)
    data = (vertices, faces, values)

    vertex_colors = np.random.random((len(vertices), 3))
    layer = Surface(data, vertex_colors=vertex_colors)
    np.testing.assert_allclose(layer.vertex_colors, vertex_colors)

    layer.vertex_colors = vertex_colors**2
    np.testing.assert_allclose(layer.vertex_colors, vertex_colors**2)


@pytest.mark.parametrize(
    ('ray_start', 'ray_direction', 'expected_value', 'expected_index'),
    [
        ([0, 1, 1], [1, 0, 0], 2, 0),
        ([10, 1, 1], [-1, 0, 0], 2, 1),
    ],
)
def test_get_value_3d(
    ray_start, ray_direction, expected_value, expected_index
):
    vertices = np.array(
        [
            [3, 0, 0],
            [3, 0, 3],
            [3, 3, 0],
            [5, 0, 0],
            [5, 0, 3],
            [5, 3, 0],
            [2, 50, 50],
            [2, 50, 100],
            [2, 100, 50],
        ]
    )
    faces = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    values = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3])
    surface_layer = Surface((vertices, faces, values))

    surface_layer._slice_dims(Dims(ndim=3, ndisplay=3))
    value, index = surface_layer.get_value(
        position=ray_start,
        view_direction=ray_direction,
        dims_displayed=[0, 1, 2],
        world=False,
    )
    assert index == expected_index
    np.testing.assert_allclose(value, expected_value)


@pytest.mark.parametrize(
    ('ray_start', 'ray_direction', 'expected_value', 'expected_index'),
    [
        ([0, 0, 1, 1], [0, 1, 0, 0], 2, 0),
        ([0, 10, 1, 1], [0, -1, 0, 0], 2, 1),
    ],
)
def test_get_value_3d_nd(
    ray_start, ray_direction, expected_value, expected_index
):
    vertices = np.array(
        [
            [0, 3, 0, 0],
            [0, 3, 0, 3],
            [0, 3, 3, 0],
            [0, 5, 0, 0],
            [0, 5, 0, 3],
            [0, 5, 3, 0],
            [0, 2, 50, 50],
            [0, 2, 50, 100],
            [0, 2, 100, 50],
        ]
    )
    faces = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    values = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3])
    surface_layer = Surface((vertices, faces, values))

    surface_layer._slice_dims(Dims(ndim=4, ndisplay=3))
    value, index = surface_layer.get_value(
        position=ray_start,
        view_direction=ray_direction,
        dims_displayed=[1, 2, 3],
        world=False,
    )
    assert index == expected_index
    np.testing.assert_allclose(value, expected_value)


def test_surface_normals():
    """Ensure that normals can be set both with dict and SurfaceNormals.

    The model should internally always use SurfaceNormals.
    """
    vertices = np.array(
        [
            [3, 0, 0],
            [3, 0, 3],
            [3, 3, 0],
            [5, 0, 0],
            [5, 0, 3],
            [5, 3, 0],
            [2, 50, 50],
            [2, 50, 100],
            [2, 100, 50],
        ]
    )
    faces = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    values = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3])

    normals = {'face': {'visible': True, 'color': 'red'}}
    surface_layer = Surface((vertices, faces, values), normals=normals)
    assert isinstance(surface_layer.normals, SurfaceNormals)
    assert surface_layer.normals.face.visible is True
    assert np.array_equal(surface_layer.normals.face.color, (1, 0, 0, 1))

    surface_layer = Surface(
        (vertices, faces, values), normals=SurfaceNormals(**normals)
    )
    assert isinstance(surface_layer.normals, SurfaceNormals)
    assert surface_layer.normals.face.visible is True
    assert np.array_equal(surface_layer.normals.face.color, (1, 0, 0, 1))


def test_surface_wireframe():
    """Ensure that wireframe can be set both with dict and SurfaceWireframe.

    The model should internally always use SurfaceWireframe.
    """
    vertices = np.array(
        [
            [3, 0, 0],
            [3, 0, 3],
            [3, 3, 0],
            [5, 0, 0],
            [5, 0, 3],
            [5, 3, 0],
            [2, 50, 50],
            [2, 50, 100],
            [2, 100, 50],
        ]
    )
    faces = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    values = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3])

    wireframe = {'visible': True, 'color': 'red'}
    surface_layer = Surface((vertices, faces, values), wireframe=wireframe)
    assert isinstance(surface_layer.wireframe, SurfaceWireframe)
    assert surface_layer.wireframe.visible is True
    assert np.array_equal(surface_layer.wireframe.color, (1, 0, 0, 1))

    surface_layer = Surface(
        (vertices, faces, values), wireframe=SurfaceWireframe(**wireframe)
    )
    assert isinstance(surface_layer.wireframe, SurfaceWireframe)
    assert surface_layer.wireframe.visible is True
    assert np.array_equal(surface_layer.wireframe.color, (1, 0, 0, 1))


def test_surface_copy():
    vertices = np.array(
        [
            [3, 0, 0],
            [3, 0, 3],
            [3, 3, 0],
            [5, 0, 0],
            [5, 0, 3],
            [5, 3, 0],
            [2, 50, 50],
            [2, 50, 100],
            [2, 100, 50],
        ]
    )
    faces = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    values = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3])

    l1 = Surface((vertices, faces, values))
    l2 = copy.copy(l1)
    assert l1.data[0] is not l2.data[0]


def test_surface_with_no_visible_faces():
    points = np.array([[0, 0.0, 0.0, 0.0], [0, 1.0, 0, 0], [0, 1, 1, 0]])
    faces = np.array([[0, 1, 2]])
    layer = Surface((points, faces))
    # the following with throw an exception when _view_faces
    # is non-integer values.
    with pytest.raises(
        ValueError, match='operands could not be broadcast together'
    ):
        layer._get_value_3d(
            np.array([1, 0, 0, 0]), np.array([1, 1, 0, 0]), [1, 2, 3]
        )


def test_docstring():
    validate_all_params_in_docstring(Surface)
    validate_kwargs_sorted(Surface)



from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest
from npe2 import DynamicPlugin, PluginManager, PluginManifest

import napari_builtins
from napari import layers


@pytest.fixture(autouse=True)
def mock_npe2_pm():
    """Mock plugin manager with no registered plugins."""
    with patch.object(PluginManager, 'discover'):
        _pm = PluginManager()
    with patch('npe2.PluginManager.instance', return_value=_pm):
        yield _pm


@pytest.fixture(autouse=True)
def use_builtins(mock_npe2_pm: PluginManager):
    plugin = DynamicPlugin('napari', plugin_manager=mock_npe2_pm)
    mf = PluginManifest.from_file(
        Path(napari_builtins.__file__).parent / 'builtins.yaml'
    )
    plugin.manifest = mf
    with plugin:
        yield plugin


LAYERS: list[layers.Layer] = [
    layers.Image(np.random.rand(10, 10)),
    layers.Labels(np.random.randint(0, 16000, (32, 32), 'uint64')),
    layers.Points(np.random.rand(20, 2)),
    layers.Points(
        np.random.rand(20, 2), properties={'values': np.random.rand(20)}
    ),
    layers.Shapes(
        [
            [(0, 0), (1, 1)],
            [(5, 7), (10, 10)],
            [(1, 3), (2, 4), (3, 5), (4, 6), (5, 7), (6, 8)],
            [(4, 3), (5, -4), (6.1, 5), (7, 6.5), (8, 7), (9, 8)],
            [(5.4, 6.7), (1.2, -3)],
        ],
        shape_type=['ellipse', 'line', 'path', 'polygon', 'rectangle'],
    ),
]


@pytest.fixture(params=LAYERS)
def some_layer(request):
    return request.param


@pytest.fixture
def layers_list():
    return LAYERS



from textwrap import dedent

import pytest

from napari.utils.tree import Group, Node


@pytest.fixture
def tree():
    return Group(
        [
            Node(name='1'),
            Group(
                [
                    Node(name='2'),
                    Group([Node(name='3'), Node(name='4')], name='g2'),
                    Node(name='5'),
                    Node(name='6'),
                    Node(name='7'),
                ],
                name='g1',
            ),
            Node(name='8'),
            Node(name='9'),
        ],
        name='root',
    )


def test_tree_str(tree):
    expected = dedent(
        """
        root
          ├──1
          ├──g1
          │  ├──2
          │  ├──g2
          │  │  ├──3
          │  │  └──4
          │  ├──5
          │  ├──6
          │  └──7
          ├──8
          └──9"""
    ).strip()
    assert str(tree) == expected


def test_node_indexing(tree: Group):
    expected_indices = [
        0,
        1,
        (1, 0),
        (1, 1),
        (1, 1, 0),
        (1, 1, 1),
        (1, 2),
        (1, 3),
        (1, 4),
        2,
        3,
    ]
    assert list(tree._iter_indices()) == expected_indices

    for index in tree._iter_indices():
        assert tree.index(tree[index]) == index

        item = tree[index]
        if item.parent:
            assert item.parent.index(item) is not None


def test_relative_node_indexing(tree):
    """Test that nodes know their index relative to parent and root."""
    root: Group[Node] = tree
    assert root.is_group()
    assert not root[0].is_group()

    assert root.index_from_root() == ()
    assert root.index_in_parent() is None
    g1 = root[1]
    assert g1.name == 'g1'
    assert g1.index_in_parent() == 1
    assert g1.index_from_root() == (1,)
    g1_1 = g1[1]
    assert g1_1.name == 'g2'
    assert g1_1.parent is g1
    assert g1_1.parent.parent is root
    assert g1_1 is tree[1, 1]  # nested index variant

    assert g1_1.index_from_root() == (1, 1)
    assert g1_1.index_in_parent() == 1
    g1_1_0 = g1_1[0]
    assert g1_1_0.index_from_root() == (1, 1, 0)
    assert g1_1_0.index_in_parent() == 0
    assert g1_1_0.name == '3'
    assert g1_1_0 is tree[1, 1, 0]  # nested index variant

    g1_1_0.unparent()
    assert g1_1_0.index_from_root() == ()
    assert g1_1_0.index_in_parent() is None

    with pytest.raises(IndexError) as e:
        g1_1_0.unparent()
    assert 'Cannot unparent orphaned Node' in str(e)


def test_traverse(tree):
    """Test depth first traversal."""
    # iterating a group just returns its children
    assert [x.name for x in tree] == ['1', 'g1', '8', '9']
    # traversing a group does a depth first traversal, including both groups
    # and nodes
    names = [x.name for x in tree.traverse()]
    e = ['root', '1', 'g1', '2', 'g2', '3', '4', '5', '6', '7', '8', '9']
    assert names == e

    # traversing leaves_only=True returns only the Nodes, not the Groups
    names = [x.name for x in tree.traverse(leaves_only=True)]
    e = ['1', '2', '3', '4', '5', '6', '7', '8', '9']
    assert names == e

    assert tree.is_group()
    g1 = tree[1]
    assert g1.parent is tree
    assert g1.name == 'g1'
    assert g1.is_group()
    g2 = g1[1]
    assert g2.parent is g1
    assert g2.name == 'g2'
    assert g2.is_group()


def test_slicing(tree):
    """Indexing into a group returns a group instance."""
    assert tree.is_group()
    slc = tree[::-2]  # take every other item, starting from the end
    assert [x.name for x in slc] == ['9', 'g1']
    assert slc.is_group()
    expected = ['Group', '9', 'g1', '2', 'g2', '3', '4', '5', '6', '7']
    assert [x.name for x in slc.traverse()] == expected


def test_contains(tree):
    """Test that the ``in`` operator works for nested nodes."""
    g1 = tree[1]
    assert g1.name == 'g1'
    assert g1 in tree

    g1_0 = g1[0]
    assert g1_0.name == '2'
    assert g1_0 in g1
    assert g1_0 in tree

    # If you need to know if an item is an immediate child, you can use parent
    assert g1.parent is tree
    assert g1_0.parent is g1

    g2 = g1[1]
    assert g2.name == 'g2'
    assert g2.is_group()
    assert g2 in tree

    g2_0 = g2[0]
    assert g2_0.name == '3'


def test_deletion(tree):
    """Test that deletion removes parent"""
    g1 = tree[1]  # first group in tree
    assert g1.parent is tree
    assert g1 in tree
    n1 = g1[0]  # first item in group1

    del tree[1]  # delete g1 from the tree
    assert g1.parent is not tree
    assert g1 not in tree
    # the tree no longer has g1 or any of its children
    assert [x.name for x in tree.traverse()] == ['root', '1', '8', '9']

    # g1 remains intact
    expected = ['g1', '2', 'g2', '3', '4', '5', '6', '7']
    assert [x.name for x in g1.traverse()] == expected
    expected = ['2', '3', '4', '5', '6', '7']
    assert [x.name for x in g1.traverse(leaves_only=True)] == expected

    # we can also delete slices, including extended slices
    del g1[1::2]
    assert n1.parent is g1  # the g1 tree is still intact
    assert [x.name for x in g1.traverse()] == ['g1', '2', '5', '7']


def test_nested_deletion(tree):
    """Test that we can delete nested indices from the root."""
    # a tree is a NestedEventedList, so we can use nested_indices
    node5 = tree[1, 2]
    assert node5.name == '5'
    del tree[1, 2]
    assert node5 not in tree

    # nested indices may also be slices
    g2 = tree[1, 1]
    node4 = g2[1]
    assert node4 in tree
    del tree[1, 1, :]  # delete all members of g2 inside of tree
    assert node4 not in tree  # node4 is gone
    assert g2 == []
    assert g2 in tree  # the group itself remains in the tree


def test_deep_index(tree: Group):
    """Test deep indexing"""

    node = tree[(1, 0)]
    assert tree.index(node) == (1, 0)


def test_remove_selected(tree: Group):
    """Test remove_selected works, with nested"""
    node = tree[(1, 0)]
    tree.selection.active = node
    tree.remove_selected()


def test_nested_custom_lookup(tree: Group):
    tree._lookup = {str: lambda x: x.name}

    # first level
    g1 = tree[1]
    assert g1.name == 'g1'  # index with integer as usual
    assert tree.index('g1') == 1
    assert tree['g1'] == g1  # index with string also works

    # second level
    g1_2 = g1[2]
    assert tree[1, 2].name == '5'
    assert tree.index('5') == (1, 2)
    assert tree['5'] == g1_2



"""
These convenience functions will be useful for searching pypi for packages
that match the plugin naming convention, and retrieving related metadata.
"""

import json
from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from typing import (
    TypedDict,
    cast,
)
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from npe2 import PackageMetadata
from typing_extensions import NotRequired

from napari.plugins.utils import normalized_name
from napari.utils.notifications import show_warning

PyPIname = str


@lru_cache
def _user_agent() -> str:
    """Return a user agent string for use in http requests."""
    import platform

    from napari import __version__
    from napari.utils import misc

    if misc.running_as_constructor_app():
        env = 'constructor'
    elif misc.in_jupyter():
        env = 'jupyter'
    elif misc.in_ipython():
        env = 'ipython'
    else:
        env = 'python'

    parts = [
        ('napari', __version__),
        ('runtime', env),
        (platform.python_implementation(), platform.python_version()),
        (platform.system(), platform.release()),
    ]
    return ' '.join(f'{k}/{v}' for k, v in parts)


class _ShortSummaryDict(TypedDict):
    """Objects returned at https://npe2api.vercel.app/api/extended_summary ."""

    name: NotRequired[PyPIname]
    version: str
    summary: str
    author: str
    license: str
    home_page: str


class SummaryDict(_ShortSummaryDict):
    display_name: NotRequired[str]
    pypi_versions: NotRequired[list[str]]
    conda_versions: NotRequired[list[str]]


def plugin_summaries() -> list[SummaryDict]:
    """Return PackageMetadata object for all known napari plugins."""
    url = 'https://npe2api.vercel.app/api/extended_summary'
    with urlopen(Request(url, headers={'User-Agent': _user_agent()})) as resp:
        return json.load(resp)


@lru_cache
def conda_map() -> dict[PyPIname, str | None]:
    """Return map of PyPI package name to conda_channel/package_name ()."""
    url = 'https://npe2api.vercel.app/api/conda'
    with urlopen(Request(url, headers={'User-Agent': _user_agent()})) as resp:
        return json.load(resp)


def iter_napari_plugin_info() -> Iterator[tuple[PackageMetadata, bool, dict]]:
    """Iterator of tuples of ProjectInfo, Conda availability for all napari plugins."""
    try:
        with ThreadPoolExecutor() as executor:
            data = executor.submit(plugin_summaries)
            _conda = executor.submit(conda_map)

        conda = _conda.result()
        data_set = data.result()
    except (HTTPError, URLError):
        show_warning(
            'There seems to be an issue with network connectivity. '
            'Remote plugins cannot be installed, only local ones.\n'
        )
        return

    conda_set = {normalized_name(x) for x in conda}
    for info in data_set:
        info_copy = dict(info)
        info_copy.pop('display_name', None)
        pypi_versions = info_copy.pop('pypi_versions')
        conda_versions = info_copy.pop('conda_versions')
        info_ = cast(_ShortSummaryDict, info_copy)

        # TODO: use this better.
        # this would require changing the api that qt_plugin_dialog expects to
        # receive

        # TODO: once the new version of npe2 is out, this can be refactored
        # to all the metadata includes the conda and pypi versions.
        extra_info = {
            'home_page': info_.get('home_page', ''),
            'display_name': info.get('display_name', ''),
            'pypi_versions': pypi_versions,
            'conda_versions': conda_versions,
        }
        info_['name'] = normalized_name(info_['name'])
        meta = PackageMetadata(**info_)  # type:ignore[call-arg]

        yield meta, (info_['name'] in conda_set), extra_info



import re
from typing import Union

import numpy as np
from qtpy.QtCore import QEvent, Qt, Signal, Slot
from qtpy.QtGui import QColor, QKeyEvent, QMouseEvent
from qtpy.QtWidgets import (
    QColorDialog,
    QCompleter,
    QFrame,
    QHBoxLayout,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)
from vispy.color import get_color_dict

from napari._qt.dialogs.qt_modal import QtPopup
from napari.utils.colormaps.colormap_utils import ColorType
from napari.utils.colormaps.standardize_color import (
    hex_to_name,
    rgb_to_hex,
    transform_color,
)
from napari.utils.translations import trans

# matches any 3- or 4-tuple of int or float, with or without parens
# captures the numbers into groups.
# this is used to allow users to enter colors as e.g.: "(1, 0.7, 0)"
rgba_regex = re.compile(
    r'\(?([\d.]+),\s*([\d.]+),\s*([\d.]+),?\s*([\d.]+)?\)?'
)

TRANSPARENT = np.array([0, 0, 0, 0], np.float32)
AnyColorType = Union[ColorType, QColor]


class QColorSwatchEdit(QWidget):
    """A widget that combines a QColorSwatch with a QColorLineEdit.

    emits a color_changed event with a 1x4 numpy array when the current color
    changes.  Note, the "model" for the current color is the ``_color``
    attribute on the QColorSwatch.

    Parameters
    ----------
    parent : QWidget, optional
        parent widget, by default None
    initial_color : AnyColorType, optional
        Starting color, by default None
    tooltip : str, optional
        Tooltip when hovering on the swatch,
        by default 'click to set color'

    Attributes
    ----------
    line_edit : QColorLineEdit
        An instance of QColorLineEdit, which takes hex, rgb, or autocompletes
        common color names.  On invalid input, this field will return to the
        previous color value.
    color_swatch : QColorSwatch
        The square that shows the current color, and can be clicked to show a
        color dialog.
    color : np.ndarray
        The current color (just an alias for the colorSwatch.color)

    Signals
    -------
    color_changed : np.ndarray
        Emits the new color when the current color changes.
    """

    color_changed = Signal(np.ndarray)

    def __init__(
        self,
        parent: QWidget | None = None,
        *,
        initial_color: AnyColorType | None = None,
        tooltip: str | None = None,
    ) -> None:
        super().__init__(parent=parent)
        self.setObjectName('QColorSwatchEdit')

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(6)
        self.setLayout(layout)

        self.line_edit = QColorLineEdit(self)
        self.line_edit.editingFinished.connect(self._on_line_edit_edited)

        self.color_swatch = QColorSwatch(self, tooltip=tooltip)
        self.color_swatch.color_changed.connect(self._on_swatch_changed)
        self.setColor = self.color_swatch.setColor
        if initial_color is not None:
            self.setColor(initial_color)

        layout.addWidget(self.color_swatch)
        layout.addWidget(self.line_edit)

    @property
    def color(self):
        """Return the current color."""
        return self.color_swatch.color

    def _on_line_edit_edited(self):
        """When the user hits enter or loses focus on the LineEdit widget."""
        text = self.line_edit.text()
        rgb_match = rgba_regex.match(text)
        if rgb_match:
            text = [float(x) for x in rgb_match.groups() if x]
        self.color_swatch.setColor(text)

    @Slot(np.ndarray)
    def _on_swatch_changed(self, color: np.ndarray):
        """Receive QColorSwatch change event, update the lineEdit, re-emit."""
        self.line_edit.setText(color)
        self.color_changed.emit(color)


class QColorSwatch(QFrame):
    """A QFrame that displays a color and can be clicked to show a QColorPopup.

    Parameters
    ----------
    parent : QWidget, optional
        parent widget, by default None
    tooltip : Optional[str], optional
        Tooltip when hovering on swatch,
        by default 'click to set color'
    initial_color : ColorType, optional
        initial color, by default will be transparent

    Attributes
    ----------
    color : np.ndarray
        The current color

    Signals
    -------
    color_changed : np.ndarray
        Emits the new color when the current color changes.
    """

    color_changed = Signal(np.ndarray)

    def __init__(
        self,
        parent: QWidget | None = None,
        tooltip: str | None = None,
        initial_color: ColorType | None = None,
    ) -> None:
        super().__init__(parent)
        self.setObjectName('colorSwatch')
        self.setToolTip(tooltip or trans._('click to set color'))
        self.setCursor(Qt.CursorShape.PointingHandCursor)

        self.color_changed.connect(self._update_swatch_style)
        self._color: np.ndarray = TRANSPARENT
        if initial_color is not None:
            self.setColor(initial_color)

    @property
    def color(self):
        """Return the current color"""
        return self._color

    @Slot(np.ndarray)
    def _update_swatch_style(self, color: np.ndarray) -> None:
        """Convert the current color to rgba() string and update appearance."""
        rgba = f'rgba({",".join(str(int(x * 255)) for x in self._color)})'
        self.setStyleSheet('#colorSwatch {background-color: ' + rgba + ';}')

    def mouseReleaseEvent(self, event: QMouseEvent):
        """Show QColorPopup picker when the user clicks on the swatch."""
        if event.button() == Qt.MouseButton.LeftButton:
            initial = QColor(*(255 * self._color).astype('int'))
            popup = QColorPopup(self, initial)
            popup.colorSelected.connect(self.setColor)
            popup.show_right_of_mouse()

    def setColor(self, color: AnyColorType) -> None:
        """Set the color of the swatch.

        Parameters
        ----------
        color : ColorType
            Can be any ColorType recognized by our
            utils.colormaps.standardize_color.transform_color function.
        """
        if isinstance(color, QColor):
            _color = (np.array(color.getRgb()) / 255).astype(np.float32)
        else:
            try:
                _color = transform_color(color)[0]
            except ValueError:
                return self.color_changed.emit(self._color)
        emit = np.any(self._color != _color)
        self._color = _color
        if emit or np.array_equiv(_color, TRANSPARENT):
            self.color_changed.emit(_color)
            return None
        return None


class QColorLineEdit(QLineEdit):
    """A LineEdit that takes hex, rgb, or autocompletes common color names.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget, by default None
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._compl = QCompleter([*get_color_dict(), 'transparent'])
        self._compl.setCompletionMode(QCompleter.InlineCompletion)
        self.setCompleter(self._compl)
        self.setTextMargins(2, 2, 2, 2)

    def setText(self, color: ColorType):
        """Set the text of the lineEdit using any ColorType.

        Colors will be converted to standard SVG spec names if possible,
        or shown as #RGBA hex if not.

        Parameters
        ----------
        color : ColorType
            Can be any ColorType recognized by our
            utils.colormaps.standardize_color.transform_color function.
        """
        _rgb = transform_color(color)[0]
        _hex = rgb_to_hex(_rgb)[0]
        super().setText(hex_to_name.get(_hex, _hex))


class CustomColorDialog(QColorDialog):
    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setObjectName('CustomColorDialog')

    def keyPressEvent(self, event: QEvent):
        event.ignore()


class QColorPopup(QtPopup):
    """A QColorDialog inside of our QtPopup.

    Allows all of the show methods of QtPopup (like show relative to mouse).
    Passes through signals from the ColorDialogm, and handles some keypress
    events.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget. by default None
    initial_color : AnyColorType, optional
        The initial color set in the color dialog, by default None

    Attributes
    ----------
    color_dialog : CustomColorDialog
        The main color dialog in the popup
    """

    currentColorChanged = Signal(QColor)
    colorSelected = Signal(QColor)

    def __init__(
        self, parent: QWidget = None, initial_color: AnyColorType = None
    ) -> None:
        super().__init__(parent)
        self.setObjectName('QtColorPopup')
        self.color_dialog = CustomColorDialog(self)

        # native dialog doesn't get added to the QtPopup frame
        # so more would need to be done to use it
        self.color_dialog.setOptions(
            QColorDialog.DontUseNativeDialog | QColorDialog.ShowAlphaChannel
        )
        layout = QVBoxLayout()
        self.frame.setLayout(layout)
        layout.addWidget(self.color_dialog)

        self.color_dialog.currentColorChanged.connect(
            self.currentColorChanged.emit
        )
        self.color_dialog.colorSelected.connect(self._on_color_selected)
        self.color_dialog.rejected.connect(self._on_rejected)
        self.color_dialog.setCurrentColor(QColor(initial_color))

    def _on_color_selected(self, color: QColor):
        """When a color has beeen selected and the OK button clicked."""
        self.colorSelected.emit(color)
        self.close()

    def _on_rejected(self):
        self.close()

    def keyPressEvent(self, event: QKeyEvent):
        """Accept current color on enter, cancel on escape.

        Parameters
        ----------
        event : QKeyEvent
            The keypress event that triggered this method.
        """
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            return self.color_dialog.accept()
        if event.key() == Qt.Key.Key_Escape:
            return self.color_dialog.reject()
        self.color_dialog.keyPressEvent(event)
        return None



import numpy as np
import pytest

from napari.layers import Labels
from napari.layers.labels._labels_key_bindings import (
    new_label,
    swap_selected_and_background_labels,
)


@pytest.fixture
def labels_data_4d():
    labels = np.zeros((5, 7, 8, 9), dtype=int)
    labels[1, 2:4, 4:6, 4:6] = 1
    labels[1, 3:5, 5:7, 6:8] = 2
    labels[2, 3:5, 5:7, 6:8] = 3
    return labels


def test_max_label(labels_data_4d):
    labels = Labels(labels_data_4d)
    new_label(labels)
    assert labels.selected_label == 4


def test_swap_background_label(labels_data_4d):
    labels = Labels(labels_data_4d)
    labels.selected_label = 10
    swap_selected_and_background_labels(labels)
    assert labels.selected_label == labels.colormap.background_value
    swap_selected_and_background_labels(labels)
    assert labels.selected_label == 10



"""Module providing a unified interface for triangulation helper functions.

We don't want numba to be a required dependency. Therefore, for
numba-accelerated functions, we provide slower NumPy-only alternatives.

With this module, downstream modules can import these helper functions without
knowing which implementation is being used.
"""

from types import ModuleType

import numpy as np

from napari.layers.shapes import _accelerated_triangulate_python

_accelerated_triangulate_numba: ModuleType | None

try:
    from napari.layers.shapes import _accelerated_triangulate_numba
except ImportError:
    _accelerated_triangulate_numba = None

USE_NUMBA_FOR_EDGE_TRIANGULATION = _accelerated_triangulate_numba is not None
RUN_WARMUP = _accelerated_triangulate_numba is not None
CACHE_WARMUP = False
# Some numba functions are used even if for non-numba triangulation backends,
# *if* numba is available. This tracks whether those functions have been warmed
# up already.
UNIVERSAL_CACHE_WARMUP = False

normalize_vertices_and_edges = (
    _accelerated_triangulate_python.normalize_vertices_and_edges_py
)


ALWAYS_NUMBA = ('remove_path_duplicates', 'create_box_from_bounding')
SWAPPABLE_NUMBA = (
    'generate_2D_edge_meshes',
    'is_convex',
    'normalize_vertices_and_edges',
    'reconstruct_polygons_from_edges',
)

if _accelerated_triangulate_numba is not None:
    remove_path_duplicates = (
        _accelerated_triangulate_numba.remove_path_duplicates
    )
    create_box_from_bounding = (
        _accelerated_triangulate_numba.create_box_from_bounding
    )
    generate_2D_edge_meshes = (
        _accelerated_triangulate_numba.generate_2D_edge_meshes
    )
    is_convex = _accelerated_triangulate_numba.is_convex
    normalize_vertices_and_edges = (
        _accelerated_triangulate_numba.normalize_vertices_and_edges
    )
    reconstruct_polygons_from_edges = (
        _accelerated_triangulate_numba.reconstruct_polygons_from_edges
    )

else:
    remove_path_duplicates = (
        _accelerated_triangulate_python.remove_path_duplicates_py
    )
    create_box_from_bounding = (
        _accelerated_triangulate_python.create_box_from_bounding_py
    )
    generate_2D_edge_meshes = (
        _accelerated_triangulate_python.generate_2D_edge_meshes_py
    )
    is_convex = _accelerated_triangulate_python.is_convex_py
    reconstruct_polygons_from_edges = (
        _accelerated_triangulate_python.reconstruct_polygons_from_edges_py
    )


def _set_numba(value: bool) -> None:
    """Set the Numba backend to use.

    Parameters
    ----------
    value : bool
        If True, use the Numba backend. If False, use the pure Python backend.
    """
    global USE_NUMBA_FOR_EDGE_TRIANGULATION

    val = value and (_accelerated_triangulate_numba is not None)
    if val:
        for name in SWAPPABLE_NUMBA:
            globals()[name] = getattr(_accelerated_triangulate_numba, name)
    else:
        for name in SWAPPABLE_NUMBA:
            globals()[name] = getattr(
                _accelerated_triangulate_python, f'{name}_py'
            )

    USE_NUMBA_FOR_EDGE_TRIANGULATION = val


def _set_warmup(value: bool) -> None:
    global RUN_WARMUP
    RUN_WARMUP = value


def warmup_universal_numba() -> None:
    """Warm up the functions that are used even with compiled backends."""
    if _accelerated_triangulate_numba is None:
        # no numba, nothing to warm up
        return
    global UNIVERSAL_CACHE_WARMUP
    if UNIVERSAL_CACHE_WARMUP:
        return
    UNIVERSAL_CACHE_WARMUP = True

    for order in ('C', 'F'):
        data = np.array(
            [[0, 0], [1, 1], [0, 1], [1, 0]], dtype=np.float32, order=order
        )
        data2 = np.array([[1, 1], [10, 15]], dtype=np.float32, order=order)

        _accelerated_triangulate_numba.remove_path_duplicates(data, False)
        _accelerated_triangulate_numba.remove_path_duplicates(data, True)
        _accelerated_triangulate_numba.create_box_from_bounding(data2)


def warmup_numba_cache() -> None:
    if _accelerated_triangulate_numba is None:
        # no numba, nothing to warm up
        return
    if not RUN_WARMUP:
        warmup_universal_numba()
        return
    global CACHE_WARMUP
    if CACHE_WARMUP:
        return

    CACHE_WARMUP = True
    warmup_universal_numba()

    for order in ('C', 'F'):
        data = np.array(
            [[0, 0], [1, 1], [0, 1], [1, 0]], dtype=np.float32, order=order
        )

        if not USE_NUMBA_FOR_EDGE_TRIANGULATION:
            _accelerated_triangulate_numba.generate_2D_edge_meshes(data, True)
            _accelerated_triangulate_numba.generate_2D_edge_meshes(data, False)
            _accelerated_triangulate_numba.remove_path_duplicates(data, False)
            _accelerated_triangulate_numba.is_convex(data)
            v, e = _accelerated_triangulate_numba.normalize_vertices_and_edges(
                data
            )
            _accelerated_triangulate_numba.reconstruct_polygons_from_edges(
                v, e
            )



import numpy as np
import pytest

from napari._qt.layer_controls.qt_points_controls import QtPointsControls
from napari.layers import Points


def test_out_of_slice_display_checkbox(qtbot):
    """Changing the model attribute should update the view"""
    layer = Points(np.random.rand(10, 2))
    qtctrl = QtPointsControls(layer)
    qtbot.addWidget(qtctrl)
    combo = qtctrl.outOfSliceCheckBox

    assert layer.out_of_slice_display is False
    combo.setChecked(True)
    assert layer.out_of_slice_display is True


def test_current_size_display_in_range(qtbot):
    """Changing the model attribute should update the view"""
    layer = Points(np.random.rand(10, 2))
    qtctrl = QtPointsControls(layer)
    qtbot.addWidget(qtctrl)
    slider = qtctrl.sizeSlider
    slider.setValue(10)

    # Initial values
    assert slider.maximum() == 100
    assert slider.minimum() == 1
    assert slider.value() == 10
    assert layer.current_size == 10

    # Size event needs to be triggered manually, because no points are selected.
    layer.current_size = 5
    layer.events.size()
    assert slider.maximum() == 100
    assert slider.minimum() == 1
    assert slider.value() == 5
    assert layer.current_size == 5

    # Size event needs to be triggered manually, because no points are selected.
    layer.current_size = 100
    layer.events.size()
    assert slider.maximum() == 100
    assert slider.minimum() == 1
    assert slider.value() == 100
    assert layer.current_size == 100

    # Size event needs to be triggered manually, because no points are selected.
    layer.current_size = 200
    layer.events.size()
    assert slider.maximum() == 201
    assert slider.minimum() == 1
    assert slider.value() == 200
    assert layer.current_size == 200

    # Size event needs to be triggered manually, because no points are selected.
    with pytest.raises(ValueError, match='must be positive'):
        layer.current_size = -1000
    layer.events.size()
    assert slider.maximum() == 201
    assert slider.minimum() == 1
    assert slider.value() == 200
    assert layer.current_size == 200

    layer.current_size = 20
    layer.events.size()
    assert slider.maximum() == 201
    assert slider.minimum() == 1
    assert slider.value() == 20
    assert layer.current_size == 20

    layer.current_size = 10
    layer.events.size()
    assert slider.maximum() == 201
    assert slider.minimum() == 1
    assert slider.value() == 10
    assert layer.current_size == 10


def test_current_size_slider_properly_initialized(qtbot):
    """Changing the model attribute should update the view"""
    layer = Points(np.random.rand(10, 2), size=np.linspace(-2, 200, 10))
    qtctrl = QtPointsControls(layer)
    qtbot.addWidget(qtctrl)
    slider = qtctrl.sizeSlider
    assert slider.maximum() == 201
    assert slider.minimum() == 1
    assert slider.value() == 10
    assert layer.current_size == 10

    layer = Points(np.random.rand(10, 2), size=np.linspace(-2, 50, 10))
    qtctrl = QtPointsControls(layer)
    qtbot.addWidget(qtctrl)
    slider = qtctrl.sizeSlider
    assert slider.maximum() == 100
    assert slider.minimum() == 1
    assert slider.value() == 10
    assert layer.current_size == 10


def test_size_slider_represents_current_size(qtbot):
    """Changing the current_size attribute should update the slider"""
    layer = Points(np.random.rand(10, 2))
    qtctrl = QtPointsControls(layer)
    qtbot.addWidget(qtctrl)
    slider = qtctrl.sizeSlider
    slider.setValue(10)

    # Initial value
    assert slider.value() == 10
    assert layer.current_size == 10

    # Size event needs to be triggered manually, because no points are selected.
    layer.current_size = 5
    layer.events.current_size()
    assert slider.value() == 5
    assert layer.current_size == 5



import os
import struct
import warnings

import numpy as np

from napari._version import __version__
from napari.utils.notifications import show_warning
from napari.utils.translations import trans


def imsave(filename: str, data: 'np.ndarray'):
    """Custom implementation of imsave to avoid skimage dependency.

    Parameters
    ----------
    filename : string
        The path to write the file to.
    data : np.ndarray
        The image data.
    """
    ext = os.path.splitext(filename)[1].lower()
    # If no file extension was specified, choose .png by default
    if ext == '':
        if (
            data.ndim == 2 or (data.ndim == 3 and data.shape[-1] in {3, 4})
        ) and not np.issubdtype(data.dtype, np.floating):
            ext = '.png'
        else:
            ext = '.tif'
            filename = filename + ext
    # not all file types can handle float data
    if ext not in [
        '.tif',
        '.tiff',
        '.bsdf',
        '.im',
        '.lsm',
        '.npz',
        '.stk',
    ] and np.issubdtype(data.dtype, np.floating):
        show_warning(
            trans._(
                'Image was not saved, because image data is of dtype float.\nEither convert dtype or save as different file type (e.g. TIFF).'
            )
        )
        return
    # Save screenshot image data to output file
    if ext in ['.png']:
        imsave_png(filename, data)
    elif ext in ['.tif', '.tiff']:
        imsave_tiff(filename, data)
    else:
        import imageio.v3 as iio

        iio.imwrite(filename, data)  # for all other file extensions


def imsave_png(filename, data):
    """Save .png image to file

    PNG images created in napari have a digital watermark.
    The napari version info is embedded into the bytes of the PNG.

    Parameters
    ----------
    filename : string
        The path to write the file to.
    data : np.ndarray
        The image data.
    """
    import imageio.v3 as iio
    import PIL.PngImagePlugin

    # Digital watermark, adds info about the napari version to the bytes of the PNG file
    pnginfo = PIL.PngImagePlugin.PngInfo()
    pnginfo.add_text(
        'Software', f'napari version {__version__} https://napari.org/'
    )
    iio.imwrite(
        filename,
        data,
        extension='.png',
        plugin='pillow',
        pnginfo=pnginfo,
    )


def imsave_tiff(filename, data):
    """Save .tiff image to file

    Parameters
    ----------
    filename : string
        The path to write the file to.
    data : np.ndarray
        The image data.
    """
    import tifffile

    if data.dtype == bool:
        tifffile.imwrite(filename, data)
    else:
        try:
            tifffile.imwrite(
                filename,
                data,
                # compression arg structure since tifffile 2022.7.28
                compression='zlib',
                compressionargs={'level': 1},
            )
        except struct.error:
            # regular tiffs don't support compressed data >4GB
            # in that case a struct.error is raised, and we write with the
            # bigtiff flag. (The flag is not on by default because it is
            # not as widely supported as normal tiffs.)
            tifffile.imwrite(
                filename,
                data,
                compression='zlib',
                compressionargs={'level': 1},
                bigtiff=True,
            )


def __getattr__(name: str):
    if name in {
        'imsave_extensions',
        'write_csv',
        'read_csv',
        'csv_to_layer_data',
        'read_zarr_dataset',
    }:
        warnings.warn(
            trans._(
                '{name} was moved from napari.utils.io in v0.4.17. Import it from napari_builtins.io instead.',
                deferred=True,
                name=name,
            ),
            FutureWarning,
            stacklevel=2,
        )
        import napari_builtins.io

        return getattr(napari_builtins.io, name)

    raise AttributeError(f'module {__name__} has no attribute {name}')



import os
import runpy
import sys
from pathlib import Path

import numpy as np
import pytest
import skimage.data
from qtpy import API_NAME

import napari
from napari._qt.qt_main_window import Window
from napari.utils.notifications import notification_manager

# check if this module has been explicitly requested or `--test-examples` is included
fpath = os.path.join(*__file__.split(os.path.sep)[-4:])
if '--test-examples' not in sys.argv and fpath not in sys.argv:
    pytest.skip(
        'Use `--test-examples` to test examples.', allow_module_level=True
    )

# not testing these examples
skip = [
    '3d_kymograph_.py',  # needs tqdm, omero-py and can take some time downloading data
    'live_tiffs_.py',  # requires files
    'tiled-rendering-2d_.py',  # too slow
    'live_tiffs_generator_.py',  # to generate files for live_tiffs_.py
    'points-over-time.py',  # too resource hungry
    'embed_ipython_.py',  # fails without monkeypatch
    'new_theme.py',  # testing theme is extremely slow on CI
    'dynamic-projections-dask.py',  # extremely slow / does not finish
]
# To skip examples during docs build end name with `_.py`

# these are more interactive tools than proper examples, so skip them
# cause they are hard to adapt for testing
skip_dev = ['leaking_check.py', 'demo_shape_creation.py']

EXAMPLE_DIR = Path(__file__).parent.parent.parent.parent / 'examples/'
DEV_EXAMPLE_DIR = Path(__file__).parent.parent.parent.parent / 'examples' / 'dev'
# using f.name here and re-joining at `run_path()` for test key presentation
# (works even if the examples list is empty, as opposed to using an ids lambda)
examples = [f.name for f in EXAMPLE_DIR.glob('*.py') if f.name not in skip]
dev_examples = [f.name for f in DEV_EXAMPLE_DIR.glob('*.py') if f.name not in skip_dev]


# still some CI segfaults, but only on windows with pyqt5
if os.getenv('CI') and os.name == 'nt' and API_NAME == 'PyQt5':
    examples = []

if os.getenv('CI') and os.name == 'nt' and 'to_screenshot.py' in examples:
    examples.remove('to_screenshot.py')

@pytest.fixture
def _example_monkeypatch(monkeypatch):
    # hide viewer window
    monkeypatch.setattr(Window, 'show', lambda *a: None)
    # prevent running the event loop
    monkeypatch.setattr(napari, 'run', lambda *a, **k: None)
    # Prevent downloading example data because this sometimes fails.
    monkeypatch.setattr(
        skimage.data,
        'cells3d',
        lambda: np.zeros((60, 2, 256, 256), dtype=np.uint16),
    )

    # make sure our sys.excepthook override doesn't hide errors
    def raise_errors(etype, value, tb):
        raise value

    monkeypatch.setattr(notification_manager, 'receive_error', raise_errors)


def _run_example(example_path):
    try:
        runpy.run_path(example_path)
    except SystemExit as e:
        # we use sys.exit(0) to gracefully exit from examples
        if e.code != 0:
            raise
    finally:
        napari.Viewer.close_all()


@pytest.mark.usefixtures('_example_monkeypatch')
@pytest.mark.filterwarnings('ignore')
@pytest.mark.skipif(not examples, reason='No examples were found.')
@pytest.mark.parametrize('fname', examples)
def test_examples(builtins, fname, monkeypatch):
    """Test that all of our examples are still working without warnings."""
    example_path = str(EXAMPLE_DIR / fname)
    monkeypatch.setattr(sys, 'argv', [fname])
    _run_example(example_path)


@pytest.mark.usefixtures('_example_monkeypatch')
@pytest.mark.filterwarnings('ignore')
@pytest.mark.skipif(not dev_examples, reason='No dev examples were found.')
@pytest.mark.parametrize('fname', dev_examples)
def test_dev_examples(fname, monkeypatch):
    """Test that all of our dev examples are still working without warnings."""
    example_path = str(DEV_EXAMPLE_DIR / fname)
    monkeypatch.setattr(sys, 'argv', [fname])
    _run_example(example_path)



class Signal:
    def __init__(self):
        self.cache = {}

    def __get__(self, instance, owner):
        if instance is None:
            return self

        try:
            return self.cache[instance]
        except KeyError:
            self.cache[instance] = instance = BoundSignal()
            return instance


class BoundSignal:
    def __init__(self):
        self._subscribers = []

    def emit(self, *args):
        for sub in self._subscribers:
            sub(*args)

    def connect(self, listener):
        self._subscribers.append(listener)



import os
import sys
from inspect import Parameter, getdoc, signature

from napari.utils.migrations import rename_argument
from napari.utils.misc import camel_to_snake
from napari.utils.translations import trans

tmpl_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'add_layer.py_tmpl'
)


with open(tmpl_path) as f:
    template = f.read()


def create_func(cls, name=None, doc=None):
    """
    Creates a function (such as `add_<layer>`) to add a layer to the viewer

    The functionality is inherited from the corresponding `<layer>` class.
    """
    cls_name = cls.__name__

    if name is None:
        name = camel_to_snake(cls_name)

    if 'layer' in name:
        raise ValueError(
            trans._(
                "name {name} should not include 'layer'",
                deferred=True,
                name=name,
            )
        )

    name = 'add_' + name

    if doc is None:
        # While the original class may have Attributes in its docstring, the
        # generated function should not have an Attributes section.
        # See https://numpydoc.readthedocs.io/en/latest/format.html#documenting-classes
        doc = getdoc(cls)
        start = doc.find('\n\nParameters\n----------\n')
        end = doc.find('\n\nAttributes\n----------\n')
        if end == -1:
            end = None
        if start > 0:
            doc = doc[start:end]

        n = 'n' if cls_name[0].lower() in 'aeiou' else ''
        doc = f'Add a{n} {cls_name} layer to the layer list. ' + doc
        doc += '\n\nReturns\n-------\n'
        doc += f'layer : :class:`napari.layers.{cls_name}`'
        doc += f'\n\tThe newly-created {cls_name.lower()} layer.'
        doc = doc.expandtabs(4)

    sig = signature(cls)
    additional_parameters = []
    if hasattr(cls.__init__, '_deprecated_constructor_args'):
        additional_parameters = [
            Parameter(
                name=arg,
                kind=Parameter.KEYWORD_ONLY,
                default=None,
            )
            for arg in cls.__init__._deprecated_constructor_args
        ]
    new_sig = sig.replace(
        parameters=[
            Parameter('self', Parameter.POSITIONAL_OR_KEYWORD),
            *list(sig.parameters.values()),
            *additional_parameters,
        ],
        return_annotation=cls,
    )

    src = template.format(
        name=name,
        signature=new_sig,
        cls_name=cls_name,
    )

    execdict = {cls_name: cls, 'napari': sys.modules.get('napari')}
    code = compile(src, filename=tmpl_path, mode='exec')
    exec(code, execdict)
    func = execdict[name]

    func.__doc__ = doc
    func.__signature__ = sig.replace(
        parameters=[
            Parameter('self', Parameter.POSITIONAL_OR_KEYWORD),
            *list(sig.parameters.values()),
        ],
        return_annotation=cls,
    )

    if hasattr(cls.__init__, '_rename_argument'):
        for (
            from_name,
            to_name,
            version,
            since_version,
        ) in cls.__init__._rename_argument:
            func = rename_argument(from_name, to_name, version, since_version)(
                func
            )

    return func



import warnings
from typing import Any, TypeVar

from napari.utils.events.containers._evented_list import EventedList
from napari.utils.events.containers._nested_list import NestableEventedList
from napari.utils.events.containers._selection import Selectable
from napari.utils.translations import trans

_T = TypeVar('_T')


class SelectableEventedList(Selectable[_T], EventedList[_T]):
    """List model that also supports selection.

    Events
    ------
    inserting (index: int)
        emitted before an item is inserted at ``index``
    inserted (index: int, value: T)
        emitted after ``value`` is inserted at ``index``
    removing (index: int)
        emitted before an item is removed at ``index``
    removed (index: int, value: T)
        emitted after ``value`` is removed at ``index``
    moving (index: int, new_index: int)
        emitted before an item is moved from ``index`` to ``new_index``
    moved (index: int, new_index: int, value: T)
        emitted after ``value`` is moved from ``index`` to ``new_index``
    changed (index: int, old_value: T, value: T)
        emitted when item at ``index`` is changed from ``old_value`` to ``value``
    changed <OVERLOAD> (index: slice, old_value: List[_T], value: List[_T])
        emitted when item at ``index`` is changed from ``old_value`` to ``value``
    reordered (value: self)
        emitted when the list is reordered (eg. moved/reversed).

    selection.changed (added: Set[_T], removed: Set[_T])
        Emitted when the set changes, includes item(s) that have been added
        and/or removed from the set.
    selection.active (value: _T)
        emitted when the current item has changed.
    selection._current (value: _T)
        emitted when the current item has changed. (Private event)
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._activate_on_insert = True
        super().__init__(*args, **kwargs)
        # bound/unbound methods are ambiguous for mypy so we need to ignore
        # https://mypy.readthedocs.io/en/stable/error_code_list.html?highlight=method-assign#check-that-assignment-target-is-not-a-method-method-assign
        self.selection._pre_add_hook = self._preselect_hook

    def _preselect_hook(self, value: _T) -> _T:
        """Called before adding an item to the selection."""
        if value not in self:
            raise ValueError(
                trans._(
                    'Cannot select item that is not in list: {value!r}',
                    deferred=True,
                    value=value,
                )
            )
        return value

    def _process_delete_item(self, item: _T) -> None:
        self.selection.discard(item)

    def insert(self, index: int, value: _T) -> None:
        super().insert(index, value)
        if self._activate_on_insert:
            # Make layer selected and unselect all others
            self.selection.active = value

    def select_all(self) -> None:
        """Select all items in the list."""
        self.selection.update(self)

    def remove_selected(self) -> None:
        """Remove selected items from list."""
        idx = 0
        for i in list(self.selection):
            idx = self.index(i)
            self.remove(i)
        if isinstance(idx, int):
            new = max(0, (idx - 1))
            do_add = len(self) > new
        else:
            *root, _idx = idx
            new = (*tuple(root), _idx - 1) if _idx >= 1 else tuple(root)
            do_add = len(self) > new[0]
        if do_add:
            self.selection.add(self[new])

    def move_selected(self, index: int, insert: int) -> None:
        """Reorder list by moving the item at index and inserting it
        at the insert index. If additional items are selected these will
        get inserted at the insert index too. This allows for rearranging
        the list based on dragging and dropping a selection of items, where
        index is the index of the primary item being dragged, and insert is
        the index of the drop location, and the selection indicates if
        multiple items are being dragged. If the moved layer is not selected
        select it.

        This method is deprecated. Please use layers.move_multiple
        with layers.selection instead.

        Parameters
        ----------
        index : int
            Index of primary item to be moved
        insert : int
            Index that item(s) will be inserted at
        """
        # this is just here for now to support the old layerlist API
        warnings.warn(
            trans._(
                'move_selected is deprecated since 0.4.16. Please use layers.move_multiple with layers.selection instead.',
                deferred=True,
            ),
            FutureWarning,
            stacklevel=2,
        )
        if self[index] not in self.selection:
            self.selection.select_only(self[index])
            moving = [index]
        else:
            moving = [i for i, x in enumerate(self) if x in self.selection]
        offset = insert >= index
        self.move_multiple(moving, insert + offset)

    def select_next(self, step: int = 1, shift: bool = False) -> None:
        """Selects next item from list."""
        if self.selection and self.selection._current:
            idx = self.index(self.selection._current) + step
            if len(self) > idx >= 0:
                next_layer = self[idx]
                if shift:
                    if next_layer in self.selection:
                        self.selection.remove(self.selection._current)
                        self.selection._current = next_layer
                    else:
                        self.selection.add(next_layer)
                        self.selection._current = next_layer
                else:
                    self.selection.active = next_layer
        elif len(self) > 0:
            self.selection.active = self[-1 if step > 0 else 0]

    def select_previous(self, shift: bool = False) -> None:
        """Selects previous item from list."""
        self.select_next(-1, shift=shift)


class SelectableNestableEventedList(
    SelectableEventedList[_T], NestableEventedList[_T]
):
    pass



"""Patch callables (functions and methods).

Our perfmon system using this to patch in perf_timers, but this can be used
for any type of patching. See patch_callables() below as the main entrypoint.
"""

import logging
import types
from collections.abc import Callable
from importlib import import_module
from typing import Union

from napari.utils.translations import trans

# The parent of a callable is a module or a class, class is of type "type".
CallableParent = Union[types.ModuleType, type]

# An example PatchFunction is:
# def _patch_perf_timer(parent, callable: str, label: str) -> None
PatchFunction = Callable[[CallableParent, str, str], None]


class PatchError(Exception):
    """Failed to patch target, config file error?"""


def _patch_attribute(
    module: types.ModuleType, attribute_str: str, patch_func: PatchFunction
) -> None:
    """Patch the module's callable pointed to by the attribute string.

    Parameters
    ----------
    module : types.ModuleType
        The module to patch.
    attribute_str : str
        An attribute in the module like "function" or "class.method".
    patch_func : PatchFunction
        This function is called to perform the patch.

    """
    # We expect attribute_str is <function> or <class>.<method>. We could
    # allow nested classes and functions if we wanted to extend this some.
    if attribute_str.count('.') > 1:
        raise PatchError(
            trans._(
                'Nested attribute not found: {attribute_str}',
                deferred=True,
                attribute_str=attribute_str,
            )
        )

    if '.' in attribute_str:
        # Assume attribute_str is <class>.<method>
        class_str, callable_str = attribute_str.split('.')
        try:
            parent = getattr(module, class_str)
        except AttributeError as e:
            raise PatchError(
                trans._(
                    'Module {module_name} has no attribute {attribute_str}',
                    deferred=True,
                    module_name=module.__name__,
                    attribute_str=attribute_str,
                )
            ) from e
        parent_str = class_str
    else:
        # Assume attribute_str is <function>.
        class_str = None
        parent = module
        parent_str = module.__name__
        callable_str = attribute_str

    try:
        getattr(parent, callable_str)
    except AttributeError as e:
        raise PatchError(
            trans._(
                'Parent {parent_str} has no attribute {callable_str}',
                deferred=True,
                parent_str=parent_str,
                callable_str=callable_str,
            )
        ) from e

    label = (
        callable_str if class_str is None else f'{class_str}.{callable_str}'
    )

    # Patch it with the user-provided patch_func.
    logging.getLogger('napari').info('patching %s.%s', module.__name__, label)
    patch_func(parent, callable_str, label)


def _import_module(
    target_str: str,
) -> tuple[types.ModuleType, str] | tuple[None, None]:
    """Import the module portion of this target string.

    Try importing successively longer segments of the target_str. For example:
       napari.components.experimental.chunk._loader.ChunkLoader.load_chunk
    will import:
        napari (success)
        napari.components (success)
        napari.components.experimental (success)
        napari.components.experimental.chunk (success)
        napari.components.experimental.chunk._loader (success)
        napari.components.experimental.chunk._loader.ChunkLoader (failure, not a module)

    The last one fails because ChunkLoader is a class not a module.

    Parameters
    ----------
    target_str : str
        The fully qualified callable such as "module1.module2.function".

    Returns
    -------
    Tuple[types.ModuleType, str]
        Where the module is the inner most imported module, and the string
        is the rest of target_str that was not modules.
    """
    parts = target_str.split('.')
    module = None  # Inner-most module imported so far.

    # Progressively try to import longer and longer segments of the path.
    for i in range(1, len(target_str)):
        module_path = '.'.join(parts[:i])
        try:
            module = import_module(module_path)
        except ModuleNotFoundError as e:
            if module is None:
                # The very first top-level module import failed!
                raise PatchError(
                    trans._(
                        'Module not found: {module_path}',
                        deferred=True,
                        module_path=module_path,
                    )
                ) from e

            # We successfully imported part of the target_str but then
            # we got a failure. Usually this is because we tried
            # importing a class or function. Return the inner-most
            # module we did successfully import. And return the rest of
            # the module_path we didn't use.
            attribute_str = '.'.join(parts[i - 1 :])
            return module, attribute_str
    return None, None


def patch_callables(callables: list[str], patch_func: PatchFunction) -> None:
    """Patch the given list of callables.

    Parameters
    ----------
    callables : List[str]
        Patch all of these callables (functions or methods).
    patch_func : PatchFunction
        Called on every callable to patch it.

    Notes
    -----
    The callables list should look like:
    [
        "module.module.ClassName.method_name",
        "module.function_name"
        ...
    ]

    Nested classes and methods not allowed, but support could be added.

    An example patch_func is::

        import wrapt
        def _my_patcher(parent: CallableParent, callable: str, label: str):
            @wrapt.patch_function_wrapper(parent, callable)
            def my_announcer(wrapped, instance, args, kwargs):
                print(f"Announce {label}")
                return wrapped(*args, **kwargs)
    """
    patched: set[str] = set()

    for target_str in callables:
        if target_str in patched:
            # Ignore duplicated targets in the config file.
            logging.getLogger('napari').warning(
                'skipping duplicate %s', target_str
            )
            continue

        # Patch the target and note that we did.
        try:
            module, attribute_str = _import_module(target_str)
            if module is not None and attribute_str is not None:
                _patch_attribute(module, attribute_str, patch_func)
        except PatchError:
            # We don't stop on error because if you switch around branches
            # but keep the same config file, it's easy for your config
            # file to contain targets that aren't in the code.

            # logging.exception magically logs the stack trace too!
            logging.getLogger('napari').exception(
                'Something went wrong while patching'
            )

        patched.add(target_str)



from napari_builtins.io._read import (
    csv_to_layer_data,
    imread,
    magic_imread,
    napari_get_reader,
    read_csv,
    read_zarr_dataset,
)
from napari_builtins.io._write import (
    imsave_extensions,
    napari_write_image,
    napari_write_labels,
    napari_write_points,
    napari_write_shapes,
    write_csv,
    write_layer_data_with_plugins,
)

__all__ = [
    'csv_to_layer_data',
    'imread',
    'imsave_extensions',
    'magic_imread',
    'napari_get_reader',
    'napari_write_image',
    'napari_write_labels',
    'napari_write_points',
    'napari_write_shapes',
    'read_csv',
    'read_zarr_dataset',
    'write_csv',
    'write_layer_data_with_plugins',
]



"""This module provides helper functions for autogenerating type stubs.

It is intentended to be run as a script or module as follows:

    python -m napari.utils.stubgen module.a module.b

... where `module.a` and `module.b` are the module names for which you'd
like to generate type stubs. Stubs will be written to a `.pyi` with the
same name and directory as the input module(s).

Example
-------

    python -m napari.utils.stubgen napari.view_layers

# outputs a file to: `napari/view_layers.pyi`

Note
----
If you want to limit the objects in the module for which stubs are created,
define an __all__ = [...] attribute in the module. Otherwise, all non-private
callable methods will be stubbed.

"""

import importlib
import inspect
import subprocess
import textwrap
import typing
import warnings
from collections.abc import Iterator
from types import ModuleType
from typing import Any, get_args, get_origin, get_type_hints

PYI_TEMPLATE = """
# THIS FILE IS AUTOGENERATED BY napari.utils.stubgen
# DO NOT EDIT

from typing import List, Union, Mapping, Sequence, Tuple, Dict, Set, Any
{imports}

{body}
"""


def _guess_exports(module, exclude=()) -> list[str]:
    """If __all__ wasn't provided, this function guesses what to stub."""
    return [
        k
        for k, v in vars(module).items()
        if callable(v) and not k.startswith('_') and k not in exclude
    ]


def _iter_imports(hint) -> Iterator[str]:
    """Get all imports necessary for `hint`"""
    # inspect.formatannotation strips "typing." from type annotations
    # so our signatures won't have it in there
    if not repr(hint).startswith('typing.') and (orig := get_origin(hint)):
        yield orig.__module__

    for arg in get_args(hint):
        yield from _iter_imports(arg)

    if isinstance(hint, list):
        for i in hint:
            yield from _iter_imports(i)
    elif hasattr(hint, '__module__') and hint.__module__ != 'builtins':
        yield hint.__module__


def generate_function_stub(func) -> tuple[set[str], str]:
    """Generate a stub and imports for a function."""
    sig = inspect.signature(func)

    if hasattr(func, '__wrapped__'):
        # unwrap @wraps decorated functions
        func = func.__wrapped__
    globalns = {**getattr(func, '__globals__', {})}
    globalns.update(vars(typing))
    globalns.update(getattr(func, '_forwardrefns_', {}))

    hints = get_type_hints(func, globalns)
    sig = sig.replace(
        parameters=[
            p.replace(annotation=hints.get(p.name, p.empty))
            for p in sig.parameters.values()
        ],
        return_annotation=hints.get('return', inspect.Parameter.empty),
    )
    imports = set()
    for hint in hints.values():
        imports.update(set(_iter_imports(hint)))
    imports -= {'typing'}

    doc = f'"""{func.__doc__}"""' if func.__doc__ else '...'
    return imports, f'def {func.__name__}{sig}:\n    {doc}\n'


def _get_subclass_methods(cls: type[Any]) -> set[str]:
    """Return the set of method names defined (only) on a subclass."""
    all_methods = set(dir(cls))
    base_methods = (dir(base()) for base in cls.__bases__)
    return all_methods.difference(*base_methods)


def generate_class_stubs(cls: type) -> tuple[set[str], str]:
    """Generate a stub and imports for a class."""
    bases = ', '.join(f'{b.__module__}.{b.__name__}' for b in cls.__bases__)

    methods = []
    attrs = []
    imports = set()

    local_names = set(cls.__dict__).union(set(cls.__annotations__))
    for sup in cls.mro()[1:]:
        local_names.difference_update(set(sup.__dict__))

    for methname in sorted(_get_subclass_methods(cls)):
        method = getattr(cls, methname)
        if not callable(method):
            continue
        _imports, stub = generate_function_stub(method)
        imports.update(_imports)
        methods.append(stub)
    hints = get_type_hints(cls)
    for name, type_ in hints.items():
        if name not in local_names:
            continue
        if hasattr(type_, '__name__'):
            hint = f'{type_.__module__}.{type_.__name__}'
        else:
            hint = repr(type_).replace('typing.', '')
        attrs.append(f'{name}: {hint.replace("builtins.", "")}')
        imports.update(set(_iter_imports(type_)))

    doc = f'"""{cls.__doc__.lstrip()}"""' if cls.__doc__ else '...'
    stub = f'class {cls.__name__}({bases}):\n    {doc}\n'
    stub += textwrap.indent('\n'.join(attrs), '    ')
    stub += '\n' + textwrap.indent('\n'.join(methods), '    ')

    return imports, stub


def generate_module_stub(module: str | ModuleType, save=True) -> str:
    """Generate a pyi stub for a module.

    By default saves to .pyi file with the same name as the module.
    """
    if isinstance(module, str):
        module = importlib.import_module(module)

    # try to use __all__, or guess exports
    names = getattr(module, '__all__', None)
    if not names:
        names = _guess_exports(module)
        warnings.warn(
            f'Module {module.__name__} does not provide `__all__`. '
            'Guessing exports.'
        )

    # For each object, create a stub and gather imports for the top of the file
    imports = set()
    stubs = []
    for name in names:
        obj = getattr(module, name)
        if isinstance(obj, type):
            _imports, stub = generate_class_stubs(obj)
        else:
            _imports, stub = generate_function_stub(obj)
        imports.update(_imports)
        stubs.append(stub)

    # build the final file string
    importstr = '\n'.join(f'import {n}' for n in imports)
    body = '\n'.join(stubs)
    pyi = PYI_TEMPLATE.format(imports=importstr, body=body)
    # format with ruff
    pyi = pyi.replace('NoneType', 'None')

    if save:
        print(  # noqa: T201
            'Writing stub:', module.__file__.replace('.py', '.pyi')
        )
        file_path = module.__file__.replace('.py', '.pyi')
        with open(file_path, 'w') as f:
            f.write(pyi)
        subprocess.run(['ruff', 'format', file_path])
        subprocess.run(['ruff', 'check', file_path])

    return pyi


if __name__ == '__main__':
    import sys

    default_modules = ['napari.view_layers', 'napari.components.viewer_model']

    for mod in sys.argv[1:] or default_modules:
        generate_module_stub(mod)



import inspect
import time

import dask
import dask.array as da
import numpy as np
import pytest
import skimage
from hypothesis import given
from hypothesis.extra.numpy import array_shapes
from skimage.transform import pyramid_gaussian

from napari.layers.image._image_utils import guess_multiscale, guess_rgb
from napari.layers.image._slice import _ImageSliceRequest
from napari.layers.utils._slice_input import _ThickNDSlice

data_dask = da.random.random(
    size=(100_000, 1000, 1000), chunks=(1, 1000, 1000)
)


def test_guess_rgb():
    sig = inspect.signature(guess_rgb)
    min_side_len = sig.parameters['min_side_len'].default

    shape = (10, 15)  # 2D only
    assert not guess_rgb(shape)

    shape = (40, 45, 6)  # final dim is too long
    assert not guess_rgb(shape)

    shape = (min_side_len - 1, min_side_len - 1, 3)  # 2D sides too small
    assert not guess_rgb(shape)

    shape = (min_side_len - 1, min_side_len + 1, 3)  # one 2D side too small
    assert not guess_rgb(shape)

    shape = (min_side_len + 1, min_side_len + 1, 3)
    assert guess_rgb(shape)

    shape = (512, 512, 3)
    assert guess_rgb(shape)

    shape = (100, 100, 4)
    assert guess_rgb(shape)

    shape = (10, 10, 3)
    assert guess_rgb(shape, min_side_len=5)


@given(shape=array_shapes(min_dims=3, min_side=0))
def test_guess_rgb_property(shape):
    sig = inspect.signature(guess_rgb)
    min_side_len = sig.parameters['min_side_len'].default

    assert guess_rgb(shape) == (
        shape[-1] in (3, 4)
        and shape[-2] > min_side_len
        and shape[-3] > min_side_len
    )


def test_guess_multiscale():
    data = np.random.random((10, 15))
    assert not guess_multiscale(data)[0]

    data = np.random.random((10, 15, 6))
    assert not guess_multiscale(data)[0]

    data = [np.random.random((10, 15, 6))]
    assert not guess_multiscale(data)[0]

    data = [np.random.random((10, 15, 6)), np.random.random((5, 7, 3))]
    assert guess_multiscale(data)[0]

    data = [np.random.random((10, 15, 6)), np.random.random((10, 7, 3))]
    assert guess_multiscale(data)[0]

    data = tuple(data)
    assert guess_multiscale(data)[0]

    if skimage.__version__ > '0.19':
        pyramid_kwargs = {'channel_axis': None}
    else:
        pyramid_kwargs = {'multichannel': False}

    data = tuple(
        pyramid_gaussian(np.random.random((10, 15)), **pyramid_kwargs)
    )
    assert guess_multiscale(data)[0]

    data = np.asarray(
        tuple(pyramid_gaussian(np.random.random((10, 15)), **pyramid_kwargs)),
        dtype=object,
    )
    assert guess_multiscale(data)[0]

    # Check for integer overflow with big data
    s = 8192
    data = [da.ones((s,) * 3), da.ones((s // 2,) * 3), da.ones((s // 4,) * 3)]
    assert guess_multiscale(data)[0]

    # Test for overflow in calculating array sizes
    s = 17179869184
    data = [
        da.from_delayed(
            dask.delayed(lambda: None), shape=(s,) * 2, dtype=np.float64
        ),
        da.from_delayed(
            dask.delayed(lambda: None), shape=(s // 2,) * 2, dtype=np.float64
        ),
    ]
    assert guess_multiscale(data)[0]


def test_guess_multiscale_strip_single_scale():
    data = [np.empty((10, 10))]
    guess, data_out = guess_multiscale(data)
    assert data_out is data[0]
    assert guess is False


def test_guess_multiscale_non_array_list():
    """Check that non-decreasing list input raises ValueError"""
    data = [np.empty((10, 15, 6))] * 2
    with pytest.raises(ValueError, match='decreasing size'):
        _, _ = guess_multiscale(data)


def test_guess_multiscale_incorrect_order():
    data = [np.empty((10, 15)), np.empty((5, 6)), np.empty((20, 15))]
    with pytest.raises(ValueError, match='decreasing size'):
        _, _ = guess_multiscale(data)


def test_timing_multiscale_big():
    now = time.monotonic()
    assert not guess_multiscale(data_dask)[0]
    elapsed = time.monotonic() - now
    assert elapsed < 2, 'test was too slow, computation was likely not lazy'


def test_create_data_indexing():
    point = (np.nan, 10.1, 2.6, 4)
    idx = _ImageSliceRequest._point_to_slices(point)
    expected = (slice(None), 10, 3, 4)
    assert idx == expected

    # note that testing entirely out of bounds slices is wrong because these methods
    # assume the bounds check already happened
    data_slice = _ThickNDSlice(
        point=(np.nan, 10.1, 2.6, 4, -1),
        margin_left=(np.nan, 0, 1.6, 0.3, 1),
        margin_right=(np.nan, 0.1, 0.3, 0.5, 0.6),
    )
    idx = _ImageSliceRequest._data_slice_to_slices(
        data_slice, dims_displayed=(0,)
    )
    expected = (
        slice(None),
        slice(10, 11),
        slice(1, 3),
        slice(4, 5),
        slice(0, 1),
    )
    assert idx == expected



"""Triangulation utilities"""

from __future__ import annotations

from enum import Enum
from math import atan2, pi
from typing import Literal, overload

import numpy as np
from numba import njit
from numba.core import types
from numba.typed import List

from napari.layers.shapes.shape_types import (
    CoordinateArray,
    CoordinateArray2D,
    CoordinateArray3D,
    EdgeArray,
)


class Orientation(Enum):
    """Orientation of a triangle.

    The terms assume napari's preferred coordinate frame, in which
    the 0th axis, y, is pointing down, and the 1st axis, x, is pointing
    right. If one of the axes is flipped, the observed orientation
    would also flip.
    """

    clockwise = -1
    collinear = 0
    anticlockwise = 1


@njit(cache=True, inline='always')
def _dot(v0: np.ndarray, v1: np.ndarray) -> float:
    """Return the dot product of two 2D vectors.

    If the vectors have norm 1, this is the cosine of the angle between them.
    """
    return v0[0] * v1[0] + v0[1] * v1[1]


@njit(cache=True, inline='always')
def _cross_z(v0: np.ndarray, v1: np.ndarray) -> float:
    """Return the z-magnitude of the cross-product between two vectors.

    If the vectors have norm 1, this is the sine of the angle between them.
    """
    return v0[0] * v1[1] - v0[1] * v1[0]


@njit(cache=True, inline='always')
def _sign_abs(f: float) -> tuple[float, float]:
    """Return 1, -1, or 0 based on the sign of f, as well as the abs of f.

    The order of if-statements shows what the function is optimized for given
    the early returns. In this case, an array with many positive values will
    execute more quickly than one with many negative values.
    """
    if f > 0:
        return 1.0, f
    if f < 0:
        return -1.0, -f
    return 0.0, 0.0


@njit(cache=True, inline='always')
def _orthogonal_vector(v: np.ndarray, ccw: bool = True) -> np.ndarray:
    """Return an orthogonal vector to v (2D).

    Parameters
    ----------
    v : np.ndarray of float, shape (2,)
        The input vector.
    ccw : bool
        Whether you want the orthogonal vector in the counterclockwise
        direction (in the napari reference frame) or clockwise.

    Returns
    -------
    np.ndarray, shape (2,)
        A vector orthogonal to the input vector, and of the same magnitude.
    """
    if ccw:
        return np.array([v[1], -v[0]])
    return np.array([-v[1], v[0]])


@njit(cache=True, inline='always')
def _calc_output_size(
    direction_vectors: np.ndarray,
    closed: bool,
    cos_miter_limit: float,
    bevel: bool,
) -> int:
    """Calculate the size of the output arrays for the triangulation.

    Parameters
    ----------
    direction_vectors : np.ndarray
        Nx2 array of direction vectors along the path. The direction vectors
        have norm 1, so computing the cosine between them is just the dot
        product.
    closed : bool
        True if shape edge is a closed path.
    cos_miter_limit : float
        Miter limit which determines when to switch from a miter join to a
        bevel join
    bevel : bool
        If True, a bevel join is always used.
        If False, a bevel join will be used when the miter limit is exceeded.

    Returns
    -------
    int
        number of points in the output arrays

    Notes
    -----
    We use a cosine miter limit instead of maximum miter length for performance
    reasons. The cosine miter limit is related to the miter length by the
    following equation:

    .. math::

        c = \frac{1}{2 (l/2)^2} - 1 = \frac{2}{l^2} - 1
    """
    n = len(direction_vectors)
    # for every miter join, we have two points on either side of the path
    # vertex; if the path is closed, we add two more points — repeats of the
    # start of the path
    point_count = 2 * n + 2 * closed

    if bevel:
        # if every join is a bevel join, the computation is trivial — we need
        # one more point at each path vertex, removing the first and last
        # vertices if the path is not closed...
        bevel_count = n - 2 * (not closed)
        # ... and we can return early
        return point_count + bevel_count

    # Otherwise, we use the norm-1 direction vectors to quickly check the
    # cosine of each angle in the path. If the angle is too sharp, we get a
    # negative cosine greater than some limit, and we add a bevel point for
    # that position.
    bevel_count = 0
    # We are effectively doing a sliding window of three points along the path
    # of n points. If the path is closed, we start with indices (-1, 0, 1) and
    # end with indices (-2, -1, 0). In contrast, in an open path, we start with
    # (0, 1, 2) and end with (-3, -2, -1).
    # In the following loop, i represents the center point of the sliding
    # window. Therefore, and keeping in mind that the stop value of a range in
    # Python is exclusive, if closed we want i in the range [0, n), while if
    # open we want i in the range [1, n-1). This can be accomplished by:
    start = 1 - closed
    stop = n - 1 + closed
    for i in range(start, stop):
        cos_angle = _dot(direction_vectors[i - 1], direction_vectors[i])
        if cos_angle < cos_miter_limit:
            bevel_count += 1

    return point_count + bevel_count


@njit(cache=True, inline='always')
def _set_centers_and_offsets(
    centers: np.ndarray,
    offsets: np.ndarray,
    triangles: np.ndarray,
    vertex: np.ndarray,
    vec1: np.ndarray,
    vec2: np.ndarray,
    half_vec1_len: float,
    half_vec2_len: float,
    j: int,
    cos_limit: float,
    always_bevel: bool,
) -> int:
    """Set the centers, offsets, and triangles for a given path and position.

    This function computes the positions of the vertices of the edge triangles
    at the given path vertex and towards the next path vertex, including, if
    needed, the miter join triangle overlapping the path vertex.

    If a miter join is needed, this function will add three triangles.
    Otherwise, it will add two triangles (dividing parallelogram of the
    next edge into two triangles).

    The positions of the triangle vertices are given as the path vertex
    (repeated once for each triangle vertex), and offset vectors from that
    vertex.

    The added triangles "optimistically" index into vertices past the vertices
    added in this iteration (indices j+3 and j+4).

    Parameters
    ----------
    centers : np.ndarray of float
        Mx2 output array of central coordinates of path triangles.
    offsets : np.ndarray of float
        Mx2 output array of the offsets from the central coordinates. Offsets
        need to be scaled by the line width and then added to the centers to
        generate the actual vertices of the triangulation.
    triangles : np.ndarray of int
        (M-2)x3 output array of the indices of the vertices that will form the
        triangles of the triangulation.
    vertex : np.ndarray
        The path vertex for which the centers, offsets and triangles are
        being calculated.
    vec1 : np.ndarray
        The norm-1 direction vector from the previous path vertex to the
        current path vertex.
    vec2 : np.ndarray
        The norm-1 direction vector from the current path vertex to the next
        path vertex.
    half_vec1_len : float
        Half the length of the segment between the previous path vertex and the
        current path vertex (used for bevel join calculation).
    half_vec2_len : float
        Half the length of the segment between the current path vertex and the
        next path vertex (used for bevel join calculation).
    j : int
        The current index in the ouput arrays.
    cos_limit : float
        Miter limit which determines when to switch from a miter join to a
        bevel join, to avoid very sharp shape angles.
    always_bevel : bool
        If True, a bevel join is always used.
        If False, a bevel join will only be used when the miter limit is
        exceeded.

    Returns
    -------
    int in {2, 3}
        The number of triangles, centers and offsets added to the arrays
    """
    cos_angle = _dot(vec1, vec2)
    sin_angle = _cross_z(vec1, vec2)
    bevel = always_bevel or cos_angle < cos_limit
    for i in range(2 + bevel):
        centers[j + i] = vertex

    if sin_angle == 0:
        # if the vectors are collinear, the miter join points are exactly
        # perpendicular to the path — we can construct this vector from vec1.
        miter = np.array([vec1[1], -vec1[0]], dtype=np.float32) * 0.5
    else:
        # otherwise, we use the line intercept theorem to calculate the miter
        # direction as (vec1 - vec2) * 0.5 — these are the
        # `miter_helper_vectors` in examples/dev/triangle_edge.py — and scale.
        # If there is a bevel join, we have to make sure that the miter points
        # to the inside of the angle, *and*, we have to make sure that it does
        # not exceed the length of either incoming edge.
        # See also:
        # https://github.com/napari/napari/pull/7268#user-content-miter
        scale_factor = 1 / sin_angle
        if bevel:
            # There is a case of bevels join, and
            # there is a need to check if the miter length is not too long.
            # For performance reasons here, the miter length is estimated
            # by the inverse of the sin of the angle between the two vectors.
            # See https://github.com/napari/napari/pull/7268#user-content-bevel-cut
            sign, mag = _sign_abs(scale_factor)
            scale_factor = sign * min(mag, half_vec1_len, half_vec2_len)
        miter = (vec1 - vec2) * 0.5 * scale_factor

    if bevel:
        # add three vertices using offset vectors orthogonal to the path as
        # well as the miter vector.
        # the order in which the vertices and triangles are added depends on
        # whether the turn is clockwise or counterclockwise.
        clockwise = sin_angle < 0
        counterclockwise = not clockwise
        invert = -1.0 if counterclockwise else 1.0
        offsets[j + counterclockwise] = invert * miter
        offsets[j + clockwise] = 0.5 * _orthogonal_vector(
            vec1, ccw=counterclockwise
        )
        offsets[j + 2] = 0.5 * _orthogonal_vector(vec2, ccw=counterclockwise)
        triangles[j] = [j, j + 1, j + 2]
        triangles[j + 1] = [j + counterclockwise, j + 2, j + 3]
        triangles[j + 2] = [j + 1 + clockwise, j + 3, j + 4]

        return 3  # bevel join added 3 triangles

    # otherwise, we just use the miter vector in either direction and add two
    # triangles
    offsets[j] = miter
    offsets[j + 1] = -miter
    triangles[j] = [j, j + 1, j + 2]
    triangles[j + 1] = [j + 1, j + 2, j + 3]
    return 2  # miter join added 2 triangles


@njit(cache=True, inline='always')
def _normalize_triangle_orientation(
    triangles: np.ndarray, centers: np.ndarray, offsets: np.ndarray
) -> None:
    """Ensure vertices of all triangles are listed in the same orientation.

    In terms of napari's preferred coordinate frame (axis 0, y, is pointing
    down, axis 1, x is pointing right), this orientation is positive for
    anti-clockwise and negative for clockwise. This function normalises all
    triangle data in-place to have positive orientation.

    The orientation is useful to check if a point is inside a triangle.

    Parameters
    ----------
    triangles : np.ndarray
        (M-2)x3 array of the indices of the vertices that will form the
        triangles of the triangulation
    centers : np.ndarray
        Mx2 array central coordinates of path triangles.
    offsets : np.ndarray
        Mx2 array of the offsets to the central coordinates. Offsets need to
        be scaled by the line width and then added to the centers to
        generate the actual vertices of the triangulation
    """
    for i in range(len(triangles)):
        ti = triangles[i]
        p1 = centers[ti[0]] + offsets[ti[0]]
        p2 = centers[ti[1]] + offsets[ti[1]]
        p3 = centers[ti[2]] + offsets[ti[2]]
        if _orientation(p1, p2, p3) == Orientation.clockwise:
            triangles[i] = [ti[2], ti[1], ti[0]]


@njit(cache=True, inline='always')
def _orientation(
    p1: np.ndarray, p2: np.ndarray, p3: np.ndarray
) -> Orientation:
    """Compute the orientation of three points.

    In terms of napari's preferred coordinate frame (axis 0, y, is pointing
    down, axis 1, x is pointing right), this orientation is positive for
    anti-clockwise and negative for clockwise.

    Parameters
    ----------
    p1, p2, p3: np.ndarray, shape (2,)
        The three points to check.

    Returns
    -------
    float
        Positive if anti-clockwise, negative if clockwise, 0 if collinear.
    """
    val = (p2[0] - p1[0]) * (p3[1] - p1[1]) - (p2[1] - p1[1]) * (p3[0] - p1[0])
    if val < 0:
        return Orientation.clockwise
    if val > 0:
        return Orientation.anticlockwise
    return Orientation.collinear


@njit(cache=True, inline='always')
def _are_polar_angles_monotonic(
    vertices: CoordinateArray2D, centroid: np.ndarray
) -> bool:
    """Check if vertices are monastical in a polar system.

    The function checks if in a polar coordinate system
    with point 0 in centroid, the angle component
    of vertices coordinate is increasing.

    This is a second part of the polygon convexity check.

    Parameters
    ----------
    vertices: np.ndarray
        Array of vertex coordinates with shape (N, 2)

    Returns
    -------
    bool
        True if the polygon is simple, False otherwise
    """
    start_angle = atan2(
        vertices[0][1] - centroid[1], vertices[0][0] - centroid[0]
    )
    prev_angle = 0.0
    for i in range(1, len(vertices)):
        angle = (
            atan2(vertices[i][1] - centroid[1], vertices[i][0] - centroid[0])
            - start_angle
        )
        if angle < 0:
            angle += 2 * pi
        if angle < prev_angle:
            return False
        prev_angle = angle
    return True


@njit(cache=True, inline='always')
def is_convex(vertices: CoordinateArray2D) -> bool:
    """Check if a polygon is convex.

    A polygon is convex when all its internal angles
    are less than or equal to 180 degrees and its edges don't
    self-intersect.

    This function determines convexity by:
    1. Checking if all non-collinear angles have the same
        orientation (clockwise or counterclockwise)
    2. Verifying that the vertices' polar angles are monotonic,
        relative to the polygon's centroid.

    - If the vertices are ordered counterclockwise, the order is
    reversed before checking.
    - Polygons with fewer than 3 vertices are not considered convex.
    - A triangle is always convex.

    Parameters
    ----------
    vertices : np.ndarray
        Array of vertex coordinates with shape (N, 2),
        where N is the number of vertices

    Returns
    -------
    bool
        True if the polygon is convex, False otherwise
    """
    if len(vertices) < 3:
        return False
    if len(vertices) == 3:
        return True
    orientation_ = Orientation.collinear
    idx = 0
    n_points = vertices.shape[0]
    orientation_set = False
    for idx in range(n_points - 2):
        p1 = vertices[idx]
        p2 = vertices[idx + 1]
        p3 = vertices[idx + 2]
        current_orientation = _orientation(p1, p2, p3)
        if current_orientation != Orientation.collinear:
            if not orientation_set:
                orientation_ = current_orientation
                orientation_set = True
            elif current_orientation != orientation_:
                return False

    if orientation_ == Orientation.collinear:
        return False

    for idx0, idx1, idx2 in [
        (n_points - 2, n_points - 1, 0),
        (n_points - 1, 0, 1),
    ]:
        triangle_orientation = _orientation(
            vertices[idx0], vertices[idx1], vertices[idx2]
        )
        if triangle_orientation not in [Orientation.collinear, orientation_]:
            return False

    centroid = np.empty(2, dtype=np.float32)
    centroid[0] = np.mean(vertices[:, 0])
    centroid[1] = np.mean(vertices[:, 1])

    if orientation_ == Orientation.anticlockwise:
        return _are_polar_angles_monotonic(vertices, centroid)

    return _are_polar_angles_monotonic(vertices[::-1], centroid)


@njit(cache=True, inline='always')
def _direction_vec_and_half_length(
    path: np.ndarray, closed: bool
) -> tuple[np.ndarray, np.ndarray]:
    """Calculate the normal vector and its half-length.

    Parameters
    ----------
    path : np.ndarray
        Mx2 array representing path for which to calculate the direction
        vectors and the length of each segment.
        This function assumes that there is no point repetition in the path:
        consecutive points should not have identical coordinates.
    closed : bool
        True if the input path is closed (forms a loop): the edge from the
        last node in the path back to the first one is automatically inserted
        if so.

    Returns
    -------
    np.ndarray
        Mx2 array representing containing normal vectors of the path.
        If closed is False, the last vector has unspecified value.
    np.ndarray
        The array of limit of length of inner vectors in bevel joins.
        To reduce the graphical artifacts, the inner vectors are limited
        to half of the length of the adjacent egdes in path.
    """
    normals = np.empty_like(path)
    vec_len_arr = np.empty((len(path)), dtype=np.float32)
    for i in range(1 - closed, len(path)):
        vec_diff = path[i] - path[i - 1]
        vec_len_arr[i - 1] = np.sqrt(vec_diff[0] ** 2 + vec_diff[1] ** 2)
        normals[i - 1] = vec_diff / vec_len_arr[i - 1]

    return normals, vec_len_arr * 0.5


@njit(cache=True, inline='always')
def _generate_2D_edge_meshes_loop(
    path: np.ndarray,
    closed: bool,
    cos_limit: float,
    bevel: bool,
    direction_vectors: np.ndarray,
    bevel_limit_array: np.ndarray,
    centers: np.ndarray,
    offsets: np.ndarray,
    triangles: np.ndarray,
) -> None:
    """Main loop for :py:func:`generate_2D_edge_meshes`.

    Parameters
    ----------
    path : np.ndarray
        Nx2 array of central coordinates of path to be triangulated
    closed : bool
        Bool which determines if the path is closed or not
    cos_limit : float
        Miter limit which determines when to switch from a miter join to a
        bevel join
    bevel : bool
        Bool which if True causes a bevel join to always be used. If False
        a bevel join will only be used when the miter limit is exceeded
    direction_vectors : np.ndarray
        Nx2 array of normal vectors of the path
    bevel_limit_array : np.ndarray
        The array of limit of length of inner vectors in bevel joins.
        To reduce the graphical artifacts, the inner vectors are limited
        to half of the length of the adjacent edges in path.
    centers : np.ndarray
        Mx2 array to put central coordinates of path triangles.
    offsets : np.ndarray
        Mx2 array to put the offsets to the central coordinates that need to
        be scaled by the line width and then added to the centers to
        generate the actual vertices of the triangulation
    triangles : np.ndarray
        (M-2)x3 array to put the indices of the vertices that will form the
        triangles of the triangulation
    """
    j = 0
    if not closed:
        centers[:2] = path[0]
        offsets[0] = 0.5 * _orthogonal_vector(direction_vectors[0], ccw=True)
        offsets[1] = -offsets[0]
        triangles[0] = [0, 1, 2]
        triangles[1] = [1, 2, 3]
        j = 2

    for i in range(1 - closed, len(direction_vectors) - 1 + closed):
        j += _set_centers_and_offsets(
            centers,
            offsets,
            triangles,
            path[i],
            direction_vectors[i - 1],
            direction_vectors[i],
            bevel_limit_array[i - 1],
            bevel_limit_array[i],
            j,
            cos_limit,
            bevel,
        )
    if closed:
        centers[j] = centers[0]
        centers[j + 1] = centers[1]
        offsets[j] = offsets[0]
        offsets[j + 1] = offsets[1]
    else:
        centers[j] = path[-1]
        centers[j + 1] = path[-1]
        offsets[j] = 0.5 * _orthogonal_vector(direction_vectors[-2])
        offsets[j + 1] = -offsets[j]


@njit(cache=True, inline='always')
def _cut_end_if_repetition(path: np.ndarray) -> np.ndarray:
    """Cut the last point of the path if it is the same as the second to last point.

    Parameters
    ----------
    path : np.ndarray
        Nx2 array of central coordinates of path to be triangulated

    Returns
    -------
    np.ndarray
        Nx2 or (N-1)x2  array of central coordinates of path to be triangulated
    """
    path_ = np.asarray(path, dtype=np.float32)
    if np.all(path_[-1] == path_[-2]):
        return path_[:-1]
    return path_


# Note: removing this decorator will double execution time.
@njit(cache=True)
def generate_2D_edge_meshes(
    path: np.ndarray,
    closed: bool = False,
    limit: float = 3.0,
    bevel: bool = False,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Determines the triangulation of a path in 2D.

    The resulting `offsets`
    can be multiplied by a `width` scalar and be added to the resulting
    `centers` to generate the vertices of the triangles for the triangulation,
    i.e. `vertices = centers + width*offsets`. By using the `centers` and
    `offsets` representation, the computed triangulation can be
    independent of the line width.

    Parameters
    ----------
    path : np.ndarray
        Nx2 array of central coordinates of path to be triangulated
    closed : bool
        Bool which determines if the path is closed or not
    limit : float
        Miter limit which determines when to switch from a miter join to a
        bevel join
    bevel : bool
        Bool which if True causes a bevel join to always be used. If False
        a bevel join will only be used when the miter limit is exceeded

    Returns
    -------
    centers : np.ndarray
        Mx2 array central coordinates of path triangles.
    offsets : np.ndarray
        Mx2 array of the offsets to the central coordinates that need to
        be scaled by the line width and then added to the centers to
        generate the actual vertices of the triangulation
    triangles : np.ndarray
        (M-2)x3 array of the indices of the vertices that will form the
        triangles of the triangulation
    """

    path = _cut_end_if_repetition(path)

    if len(path) < 2:
        centers = np.empty((4, 2), dtype=np.float32)
        centers[0] = path[0]
        centers[1] = path[0]
        centers[2] = path[0]
        centers[3] = path[0]
        triangles = np.empty((2, 3), dtype=np.int32)
        triangles[0] = [0, 1, 3]
        triangles[1] = [1, 3, 2]
        return (centers, np.zeros((4, 2), dtype=np.float32), triangles)

    # why cos_limit is calculated this way is explained in the note in
    # https://github.com/napari/napari/pull/7268#user-content-bevel-limit
    cos_limit = 1 / (2 * (limit / 2) ** 2) - 1.0

    direction_vectors, bevel_limit_array = _direction_vec_and_half_length(
        path, closed
    )

    point_count = _calc_output_size(
        direction_vectors, closed, cos_limit, bevel
    )

    centers = np.empty((point_count, 2), dtype=np.float32)
    offsets = np.empty((point_count, 2), dtype=np.float32)
    triangles = np.empty((point_count - 2, 3), dtype=np.int32)

    _generate_2D_edge_meshes_loop(
        path,
        closed,
        cos_limit,
        bevel,
        direction_vectors,
        bevel_limit_array,
        centers,
        offsets,
        triangles,
    )

    # We fix triangle orientation to speed up checks for
    # whether points are in triangle.
    # Without the fix, the code is not robust to orientation.
    _normalize_triangle_orientation(triangles, centers, offsets)

    return centers, offsets, triangles


@njit(cache=True)
def remove_path_duplicates(path: np.ndarray, closed: bool) -> np.ndarray:
    """Remove consecutive duplicates from a path.

    Parameters
    ----------
    path : np.ndarray
        Nx2 or Nx3 array of central coordinates of path to be deduplicated

    Returns
    -------
    np.ndarray
        Nx2 or Nx3 array of central coordinates of deduplicated path
    """
    if len(path) <= 2:
        # part of ugly hack to keep lasso tools working
        # should be removed after fixing the lasso tool
        return path

    dup_count = 0
    # We would like to use len(path) - 1 as the range.
    # To keep the lasso tool working, we need to allow
    # duplication of the last point.
    # If the lasso tool is refactored, update to use the preferred range.
    for i in range(len(path) - 2):
        if np.all(path[i] == path[i + 1]):
            dup_count += 1

    if closed and np.all(path[0] == path[-1]):
        dup_count += 1

    if dup_count == 0:
        return path

    target_len = len(path) - dup_count
    new_path = np.empty((target_len, path.shape[1]), dtype=path.dtype)
    new_path[0] = path[0]
    index = 0
    for i in range(1, len(path)):
        if index == target_len - 1:
            break
        if np.any(new_path[index] != path[i]):
            new_path[index + 1] = path[i]
            index += 1

    return new_path


@njit(cache=True)
def create_box_from_bounding(bounding_box: np.ndarray) -> np.ndarray:
    """Creates the axis aligned interaction box of a bounding box

    Parameters
    ----------
    bounding_box : np.ndarray
        2x2 array of the bounding box. The first row is the minimum values and
        the second row is the maximum values

    Returns
    -------
    box : np.ndarray
        9x2 array of vertices of the interaction box. The first 8 points are
        the corners and midpoints of the box in clockwise order starting in the
        upper-left corner. The last point is the center of the box
    """
    x_min = bounding_box[0, 0]
    x_max = bounding_box[1, 0]
    y_min = bounding_box[0, 1]
    y_max = bounding_box[1, 1]
    result = np.empty((9, 2), dtype=np.float32)
    result[0] = [x_min, y_min]
    result[1] = [(x_min + x_max) / 2, y_min]
    result[2] = [x_max, y_min]
    result[3] = [x_max, (y_min + y_max) / 2]
    result[4] = [x_max, y_max]
    result[5] = [(x_min + x_max) / 2, y_max]
    result[6] = [x_min, y_max]
    result[7] = [x_min, (y_min + y_max) / 2]
    result[8] = [(x_min + x_max) / 2, (y_min + y_max) / 2]
    return result


@overload
def reconstruct_polygons_from_edges(
    vertices: CoordinateArray2D, edges: EdgeArray
) -> list[CoordinateArray2D]: ...


@overload
def reconstruct_polygons_from_edges(
    vertices: CoordinateArray3D, edges: EdgeArray
) -> list[CoordinateArray3D]: ...


@njit(cache=True)
def reconstruct_polygons_from_edges(
    vertices: CoordinateArray, edges: EdgeArray
) -> list[CoordinateArray2D] | list[CoordinateArray3D]:
    """Reconstruct polygons from vertices and edges.

    This function takes the output from `normalize_vertices_and_edges` — which
    is a vertex set and a list of possibly-disjoint edges — and produces a list
    of independent polygons.

    The algorithm reconstructs sub polygons using recursion.
    Starting from the first edge, it traverses the graph until it reaches starting vertex.
    After it, the algorithm iterates until reach the first non visited edge.
    So the implementation has O(M) complexity.

    Parameters
    ----------
    vertices : np.ndarray
        Array of vertex coordinates with shape (N, 2) or (N, 3)
        Cannot contain repeated vertices.
    edges : np.ndarray
        Array of edge indices with shape (M, 2)
        Cannot contain repeated edges.


    Returns
    -------
    list of np.ndarray
        List of polygons, where each polygon is an array of vertex coordinates
    """
    n_edges = edges.shape[0]
    visited = np.zeros(n_edges, dtype=np.bool_)
    n_vertices = vertices.shape[0]

    # Create a list (indexed by vertex) containing lists of incident edge indices.
    incident = List()
    for _ in range(n_vertices):
        incident.append(List.empty_list(types.int64))

    for i in range(n_edges):
        start = edges[i, 0]
        end = edges[i, 1]
        incident[start].append(i)
        incident[end].append(i)

    # List to hold all reconstructed polygons.
    polygons = List()

    for i in range(n_edges):
        if visited[i]:
            continue
        poly: list[int] = []
        # Start a new polygon with the current edge.
        start_v = edges[i, 0]
        current_v = edges[i, 1]
        poly.append(start_v)
        poly.append(current_v)
        visited[i] = True
        closed = False

        # Walk along the polygon edges until we loop back to start_v.
        while not closed:
            found = False
            # Loop through edges incident to current_v.
            for edge_idx in incident[current_v]:
                if visited[edge_idx]:
                    continue
                a = edges[edge_idx, 0]
                b = edges[edge_idx, 1]
                # Choose the vertex that is not the current one.
                next_v = a if b == current_v else b
                visited[edge_idx] = True
                if next_v == start_v:
                    closed = True
                else:
                    poly.append(next_v)
                current_v = next_v
                found = True
                if current_v == start_v:
                    closed = True
                break  # move on as soon as we find the next edge
            if not found:
                # All polygons sent as input to this function *should* be
                # closed. However, if a user accidentally passes in an open
                # chain of vertices, without this break, we would enter an
                # infinite loop. Therefore, we leave it here for safety.
                break
        polygon = vertices[np.array(poly)]
        polygons.append(polygon)
    return polygons


@njit(cache=True)
def normalize_vertices_and_edges(
    vertices: CoordinateArray2D, close: bool = False
) -> tuple[
    CoordinateArray2D,
    np.ndarray[tuple[int, Literal[2]], np.dtype[np.int64]],
]:
    """Get a list of edges that must be in triangulation for a path or polygon.

    This function ensures that:

    - no nodes are repeated, as this can cause problems with triangulation
      algorithms.
    - edges that appear twice are discarded. This allows representation of
      polygons with holes in them.

    Parameters
    ----------
    vertices: np.ndarray[np.floating], shape (N, 2)
        The 2D coordinates of the polygon's vertices. They are expected to be
        in the order in which they appear in the polygon: that is, vertices
        that follow each other are expected to be connected to each other. The
        exception is if the same edge is visited twice (in any direction): such
        edges are removed.

        Holes are expected to be represented by a polygon embedded within the
        larger polygon but winding in the opposite direction.

    close: bool
        Whether to close the polygon or treat it as a path. Note: this argument
        has no effect if the last vertex is equal to the first one — then the
        closing is explicit.

    Returns
    -------
    new_vertices: np.ndarray[np.floating], shape (M, 2)
        Vertices with duplicate nodes removed.
    edges: np.ndarray[np.intp], shape (P, 2)
        List of edges in the polygon, expressed as an array of pairs of vertex
        indices. This is usually [(0, 1), (1, 2), ... (N-1, 0)], but edges
        that are visited twice are removed.
    """
    if (
        vertices[0, 0] == vertices[-1, 0] and vertices[0, 1] == vertices[-1, 1]
    ):  # closed polygon
        vertices = vertices[:-1]  # make closing implicit
        close = True
    # Now, we make sure the vertices are unique (repeated vertices cause
    # problems in spatial algorithms, and those problems can manifest as
    # segfaults if said algorithms are implemented in C-like languages.)
    vertex_to_idx: dict[tuple[float, float], int] = {}
    new_vertices = []
    edges: set[tuple[int, int]] = set()
    prev_idx = 0
    i = 0
    for vertex in vertices:
        vertex_t = (vertex[0], vertex[1])
        if vertex_t in vertex_to_idx:
            current_idx = vertex_to_idx[vertex_t]
        else:
            current_idx = i
            vertex_to_idx[vertex_t] = i
            new_vertices.append(vertex)
            i += 1

        if prev_idx < current_idx:
            edge = (prev_idx, current_idx)
        else:
            edge = (current_idx, prev_idx)
        if edge in edges:
            edges.remove(edge)
        else:
            edges.add(edge)
        prev_idx = current_idx

    if close:
        vertex_t = (vertices[-1, 0], vertices[-1, 1])
        idx = vertex_to_idx[vertex_t]
        edge = (0, idx)
        if edge in edges:
            edges.remove(edge)
        else:
            edges.add(edge)

    edges.remove((0, 0))
    new_vertices_array = np.empty((len(new_vertices), 2), dtype=np.float32)
    for i, vertex in enumerate(new_vertices):
        new_vertices_array[i] = vertex
    edges_array = np.array(list(edges), dtype=np.int64)
    return new_vertices_array, edges_array  # type: ignore[return-value]



from typing import TYPE_CHECKING

from vispy.scene.visuals import create_visual_node
from vispy.visuals.image import ImageVisual
from vispy.visuals.shaders import Function, FunctionChain

from napari._vispy.visuals.util import TextureMixin

if TYPE_CHECKING:
    from vispy.visuals.visual import VisualView


class LabelVisual(TextureMixin, ImageVisual):
    """Visual subclass displaying a 2D array of labels."""

    def _build_color_transform(self) -> FunctionChain:
        """Build the color transform function chain."""
        funcs = [
            Function(self._func_templates['red_to_luminance']),
            Function(self.cmap.glsl_map),
        ]

        return FunctionChain(
            funcs=funcs,
        )


BaseLabel = create_visual_node(LabelVisual)


class LabelNode(BaseLabel):  # type: ignore [valid-type,misc]
    def _compute_bounds(
        self, axis: int, view: 'VisualView'
    ) -> tuple[float, float] | None:
        if self._data is None:
            return None
        elif axis > 1:  # noqa: RET505
            return 0, 0
        else:
            return 0, self.size[axis]



from __future__ import annotations

import itertools
import typing
import warnings
from collections.abc import Iterable
from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np

from napari.components.dims import RangeTuple
from napari.layers import Layer
from napari.layers.utils.layer_utils import Extent
from napari.utils.events.containers import SelectableEventedList
from napari.utils.naming import inc_name_count
from napari.utils.translations import trans

if TYPE_CHECKING:
    from npe2.manifest.io import WriterContribution
    from typing_extensions import Self


def get_name(layer: Layer) -> str:
    """Return the name of a layer."""
    return layer.name


class LayerList(SelectableEventedList[Layer]):
    """List-like layer collection with built-in reordering and callback hooks.

    Parameters
    ----------
    data : iterable
        Iterable of napari.layer.Layer

    Events
    ~~~~~~
    inserting : (index: int)
        emitted before an item is inserted at ``index``
    inserted : (index: int, value: T)
        emitted after ``value`` is inserted at ``index``
    removing : (index: int)
        emitted before an item is removed at ``index``
    removed : (index: int, value: T)
        emitted after ``value`` is removed at ``index``
    moving : (index: int, new_index: int)
        emitted before an item is moved from ``index`` to ``new_index``
    moved : (index: int, new_index: int, value: T)
        emitted after ``value`` is moved from ``index`` to ``new_index``
    changed : (index: int, old_value: T, value: T)
        emitted when item at ``index`` is changed from ``old_value`` to
        ``value``
    changed <OVERLOAD> : (index: slice, old_value: List[_T], value: List[_T])
        emitted when items at ``index``es are changed from ``old_value`` to
        ``value``
    reordered : (value: self)
        emitted when the list is reordered (eg. moved/reversed).
    selection.events.changed : (added: Set[_T], removed: Set[_T])
        emitted when the set changes, includes item(s) that have been added
        and/or removed from the set.
    selection.events.active : (value: _T)
        emitted when the current item has changed.
    selection.events._current : (value: _T)
        emitted when the current item has changed. (Private event)

    Notes
    -----

    Note that ``changed`` events are only emitted when an element of the
    list changes, *not* when the list itself changes (for example when items
    are added or removed). For example, ``layerlist.append(layer)`` will emit
    an ``inserted`` event. ``layerlist[idx] = layer`` *will* emit a ``changed``
    event.

    However, the layerlist does not have a way of detecting when an object in
    the list is modified in-place. Therefore, although
    ``layerlist[idx].scale = [2, 1, 1]`` changes the *value* of the layer at
    position ``idx``, a ``changed`` event will not be emitted.

    Examples
    --------

    >>> import napari
    >>> from skimage.data import astronaut
    >>> viewer = napari.Viewer()
    >>> event_list = []

    Connect to the event list:

    >>> viewer.layers.events.connect(event_list.append)
    <built-in method append of list object at 0x7fc225764780>

    >>> viewer.add_image(astronaut())
    <Image layer 'Image' at 0x7f0fe7fa9e90>
    >>> viewer.add_points()
    <Points layer 'Points' at 0x7f102f962350>
    >>> viewer.layers
    [<Image layer 'Image' at 0x7f0fe7fa9e90>, <Points layer 'Points' at 0x7f102f962350>]

    Inspecting the list of events, we see:

    >>> event_list[0].type
    'inserting'
    >>> viewer.layers.pop(1)
    <Points layer 'Points' at 0x7f102f962350>
    >>> event_list[-1].type
    'removed'

    """

    def __init__(self, data=()) -> None:
        super().__init__(
            data=data,
            basetype=Layer,
            lookup={str: get_name},
        )
        self._create_contexts()

    def _create_contexts(self):
        """Create contexts to manage enabled/visible action/menu states.

        Connects LayerList and Selection[Layer] to their context keys to allow
        actions and menu items (in the GUI) to be dynamically enabled/disabled
        and visible/hidden based on the state of layers in the list.
        """

        # TODO: figure out how to move this context creation bit.
        # Ideally, the app should be aware of the layerlist, but not vice versa.
        # This could probably be done by having the layerlist emit events that
        # the app connects to, then the `_ctx` object would live on the app,
        # (not here)
        from napari._app_model.context import create_context
        from napari._app_model.context._layerlist_context import (
            LayerListContextKeys,
            LayerListSelectionContextKeys,
        )

        self._ctx = create_context(self)
        if self._ctx is not None:  # happens during Viewer type creation
            self._ctx_keys = LayerListContextKeys(self._ctx)
            self.events.inserted.connect(self._ctx_keys.update)
            self.events.removed.connect(self._ctx_keys.update)

            self._selection_ctx_keys = LayerListSelectionContextKeys(self._ctx)
            self.selection.events.changed.connect(
                self._selection_ctx_keys.update
            )

    def _process_delete_item(self, item: Layer):
        super()._process_delete_item(item)
        item.events.extent.disconnect(self._clean_cache)
        item.events._extent_augmented.disconnect(self._clean_cache)
        self._clean_cache()

    def _clean_cache(self):
        cached_properties = (
            'extent',
            '_extent_world',
            '_extent_world_augmented',
            '_step_size',
        )
        [self.__dict__.pop(p, None) for p in cached_properties]

    def __newlike__(self, data):
        return LayerList(data)

    def _coerce_name(self, name, layer=None):
        """Coerce a name into a unique equivalent.

        Parameters
        ----------
        name : str
            Original name.
        layer : napari.layers.Layer, optional
            Layer for which name is generated.

        Returns
        -------
        new_name : str
            Coerced, unique name.
        """
        existing_layers = {x.name for x in self if x is not layer}
        for _ in range(len(self)):
            if name in existing_layers:
                name = inc_name_count(name)
        return name

    def _update_name(self, event):
        """Coerce name of the layer in `event.layer`."""
        layer = event.source
        layer.name = self._coerce_name(layer.name, layer)

    def _ensure_unique(self, values, allow=()):
        bad = set(self._list) - set(allow)
        values = tuple(values) if isinstance(values, Iterable) else (values,)
        for v in values:
            if v in bad:
                raise ValueError(
                    trans._(
                        "Layer '{v}' is already present in layer list",
                        deferred=True,
                        v=v,
                    )
                )
        return values

    @typing.overload
    def __getitem__(self, item: int | str) -> Layer: ...

    @typing.overload
    def __getitem__(self, item: slice) -> Self: ...

    def __getitem__(self, item):
        return super().__getitem__(item)

    def __setitem__(self, key, value):
        old = self._list[key]
        if isinstance(key, slice):
            value = self._ensure_unique(value, old)
        elif isinstance(key, int):
            (value,) = self._ensure_unique((value,), (old,))
        super().__setitem__(key, value)

    def insert(self, index: int, value: Layer):
        """Insert ``value`` before index."""
        (value,) = self._ensure_unique((value,))
        new_layer = self._type_check(value)
        new_layer.name = self._coerce_name(new_layer.name)
        self._clean_cache()
        new_layer.events.extent.connect(self._clean_cache)
        new_layer.events._extent_augmented.connect(self._clean_cache)
        super().insert(index, new_layer)

    def remove_selected(self):
        """Remove selected layers from LayerList, but first unlink them."""
        if not self.selection:
            return
        self.unlink_layers(self.selection)
        super().remove_selected()

    def toggle_selected_visibility(self):
        """Toggle visibility of selected layers"""
        for layer in self.selection:
            layer.visible = not layer.visible

    @cached_property
    def _extent_world(self) -> np.ndarray:
        """Extent of layers in world coordinates.

        Default to 2D with (-0.5, 511.5) min/ max values if no data is present.
        Corresponds to pixels centered at [0, ..., 511].

        Returns
        -------
        extent_world : array, shape (2, D)
        """
        return self._get_extent_world([layer.extent for layer in self])

    @cached_property
    def _extent_world_augmented(self) -> np.ndarray:
        """Extent of layers in world coordinates.

        Default to 2D with (-0.5, 511.5) min/ max values if no data is present.
        Corresponds to pixels centered at [0, ..., 511].

        Returns
        -------
        extent_world : array, shape (2, D)
        """
        return self._get_extent_world(
            [layer._extent_augmented for layer in self],
            augmented=True,
        )

    def _get_min_and_max(self, mins_list, maxes_list):
        # Reverse dimensions since it is the last dimensions that are
        # displayed.
        mins_list = [mins[::-1] for mins in mins_list]
        maxes_list = [maxes[::-1] for maxes in maxes_list]

        with warnings.catch_warnings():
            # Taking the nanmin and nanmax of an axis of all nan
            # raises a warning and returns nan for that axis
            # as we have do an explicit nan_to_num below this
            # behaviour is acceptable and we can filter the
            # warning
            warnings.filterwarnings(
                'ignore',
                message=str(
                    trans._('All-NaN axis encountered', deferred=True)
                ),
            )
            min_v = np.nanmin(
                list(itertools.zip_longest(*mins_list, fillvalue=np.nan)),
                axis=1,
            )
            max_v = np.nanmax(
                list(itertools.zip_longest(*maxes_list, fillvalue=np.nan)),
                axis=1,
            )

        # 512 element default extent as documented in `_get_extent_world`
        min_v = np.nan_to_num(min_v, nan=-0.5)
        max_v = np.nan_to_num(max_v, nan=511.5)

        # switch back to original order
        return min_v[::-1], max_v[::-1]

    def _get_extent_world(self, layer_extent_list, augmented=False):
        """Extent of layers in world coordinates.

        Default to 2D image-like with (0, 511) min/ max values if no data is present.
        Corresponds to image with 512 pixels in each dimension.

        Returns
        -------
        extent_world : array, shape (2, D)
        """
        if len(self) == 0:
            min_v = np.zeros(self.ndim)
            max_v = np.full(self.ndim, 511.0)
            # image-like augmented extent is actually expanded by 0.5
            if augmented:
                min_v -= 0.5
                max_v += 0.5
        else:
            extrema = [extent.world for extent in layer_extent_list]
            mins = [e[0] for e in extrema]
            maxs = [e[1] for e in extrema]
            min_v, max_v = self._get_min_and_max(mins, maxs)

        return np.vstack([min_v, max_v])

    @cached_property
    def _step_size(self) -> np.ndarray:
        """Ideal step size between planes in world coordinates.

        Computes the best step size that allows all data planes to be
        sampled if moving through the full range of world coordinates.
        The current implementation just takes the minimum scale.

        Returns
        -------
        step_size : array, shape (D,)
        """
        return self._get_step_size([layer.extent for layer in self])

    def _step_size_from_scales(self, scales):
        # Reverse order so last axes of scale with different ndim are aligned
        scales = [scale[::-1] for scale in scales]
        full_scales = list(
            np.array(list(itertools.zip_longest(*scales, fillvalue=np.nan)))
        )
        # restore original order
        return np.nanmin(full_scales, axis=1)[::-1]

    def _get_step_size(self, layer_extent_list):
        if len(self) == 0:
            return np.ones(self.ndim)

        scales = [extent.step for extent in layer_extent_list]
        return self._step_size_from_scales(scales)

    def get_extent(self, layers: Iterable[Layer]) -> Extent:
        """
        Return extent for a given layer list.

        Extent bounds are inclusive. This function is useful for calculating the extent
        of a subset of layers when preparing and updating some supplementary layers.
        For example see the cross Vectors layer in the `multiple_viewer_widget` example.

        Parameters
        ----------
        layers : list of Layer
            list of layers for which extent should be calculated

        Returns
        -------
        extent : Extent
            extent for selected layers
        """
        extent_list = [layer.extent for layer in layers]
        return Extent(
            data=None,
            world=self._get_extent_world(extent_list),
            step=self._get_step_size(extent_list),
        )

    @cached_property
    def extent(self) -> Extent:
        """
        Extent of layers in data and world coordinates.

        Extent bounds are inclusive. See Layer.extent for a detailed explanation
        of how extents are calculated.
        """
        return self.get_extent(list(self))

    @property
    def _ranges(self) -> tuple[RangeTuple, ...]:
        """Get ranges for Dims.range in world coordinates."""
        ext = self.extent
        return tuple(
            RangeTuple(*x)
            for x in zip(ext.world[0], ext.world[1], ext.step, strict=False)
        )

    @property
    def ndim(self) -> int:
        """Maximum dimensionality of layers.

        Defaults to 2 if no data is present.

        Returns
        -------
        ndim : int
        """
        return max((layer.ndim for layer in self), default=2)

    def _link_layers(
        self,
        method: str,
        layers: Iterable[str | Layer] | None = None,
        attributes: Iterable[str] = (),
    ):
        # adding this method here allows us to emit an event when
        # layers in this group are linked/unlinked.  Which is necessary
        # for updating context
        from napari.layers.utils import _link_layers

        if layers is not None:
            layers = [self[x] if isinstance(x, str) else x for x in layers]  # type: ignore
        else:
            layers = self
        getattr(_link_layers, method)(layers, attributes)
        self.selection.events.changed(added={}, removed={})

    def link_layers(
        self,
        layers: Iterable[str | Layer] | None = None,
        attributes: Iterable[str] = (),
    ):
        """
        Links the selected layers.

        Once layers are linked, any action performed on one layer will be
        performed on all linked layers at the same time.
        """
        return self._link_layers('link_layers', layers, attributes)

    def unlink_layers(
        self,
        layers: Iterable[str | Layer] | None = None,
        attributes: Iterable[str] = (),
    ):
        """Unlinks previously linked layers.

        Changes to one of the layer's properties no longer result in the same
        changes to the previously linked layers.
        """
        return self._link_layers('unlink_layers', layers, attributes)

    def save(
        self,
        path: str,
        *,
        selected: bool = False,
        plugin: str | None = None,
        _writer: WriterContribution | None = None,
    ) -> list[str]:
        """Save all or only selected layers to a path using writer plugins.

        If ``plugin`` is not provided and only one layer is targeted, then we
        directly call the corresponding``napari_write_<layer_type>`` hook (see
        :ref:`single layer writer hookspecs <write-single-layer-hookspecs>`)
        which will loop through implementations and stop when the first one
        returns a non-``None`` result. The order in which implementations are
        called can be changed with the Plugin sorter in the GUI or with the
        corresponding hook's
        :meth:`~napari.plugins._hook_callers._HookCaller.bring_to_front`
        method.

        If ``plugin`` is not provided and multiple layers are targeted,
        then we call
        :meth:`~napari.plugins.hook_specifications.napari_get_writer` which
        loops through plugins to find the first one that knows how to handle
        the combination of layers and is able to write the file. If no plugins
        offer :meth:`~napari.plugins.hook_specifications.napari_get_writer` for
        that combination of layers then the default
        :meth:`~napari.plugins.hook_specifications.napari_get_writer` will
        create a folder and call ``napari_write_<layer_type>`` for each layer
        using the ``Layer.name`` variable to modify the path such that the
        layers are written to unique files in the folder.

        If ``plugin`` is provided and a single layer is targeted, then we
        call the ``napari_write_<layer_type>`` for that plugin, and if it fails
        we error.

        If ``plugin`` is provided and multiple layers are targeted, then
        we call we call
        :meth:`~napari.plugins.hook_specifications.napari_get_writer` for
        that plugin, and if it doesn`t return a ``WriterFunction`` we error,
        otherwise we call it and if that fails if it we error.

        Parameters
        ----------
        path : str
            A filepath, directory, or URL to open.  Extensions may be used to
            specify output format (provided a plugin is available for the
            requested format).
        selected : bool
            Optional flag to only save selected layers. False by default.
        plugin : str, optional
            Name of the plugin to use for saving. If None then all plugins
            corresponding to appropriate hook specification will be looped
            through to find the first one that can save the data.
        _writer : WriterContribution, optional
            private: npe2 specific writer override.

        Returns
        -------
        list of str
            File paths of any files that were written.
        """
        from napari.plugins.io import save_layers

        layers = (
            [x for x in self if x in self.selection]
            if selected
            else list(self)
        )

        if selected:
            msg = trans._('No layers selected', deferred=True)
        else:
            msg = trans._('No layers to save', deferred=True)

        if not layers:
            warnings.warn(msg)
            return []

        return save_layers(path, layers, plugin=plugin, _writer=_writer)



import itertools
import os
from collections.abc import Callable, Sequence
from functools import lru_cache
from types import ModuleType
from typing import (
    Literal,
    overload,
)

import numpy as np
from skimage import morphology


def always_false(*_):
    return False


class Skip:
    def __init__(
        self,
        if_in_pr: Callable[..., bool] = always_false,
        if_on_ci: Callable[..., bool] = always_false,
        always: Callable[..., bool] = always_false,
    ):
        self.func_pr = if_in_pr if 'PR' in os.environ else always_false
        self.func_ci = if_on_ci if 'CI' in os.environ else always_false
        self.func_always = always

    def __contains__(self, item):
        return (
            self.func_pr(*item)
            or self.func_ci(*item)
            or self.func_always(*item)
        )


def _generate_ball(radius: int, ndim: int) -> np.ndarray:
    """Generate a ball of given radius and dimension.

    Parameters
    ----------
    radius : int
        Radius of the ball.
    ndim : int
        Dimension of the ball.

    Returns
    -------
    ball : ndarray of uint8
        Binary array of the hyper ball.
    """

    if ndim == 2:
        return morphology.disk(radius)
    if ndim == 3:
        return morphology.ball(radius)
    shape = (2 * radius + 1,) * ndim
    radius_sq = radius**2
    coords = np.indices(shape) - radius
    return (np.sum(coords**2, axis=0) <= radius_sq).astype(np.uint8)


def _generate_density(radius: int, ndim: int) -> np.ndarray:
    """Generate gaussian density of given radius and dimension."""
    shape = (2 * radius + 1,) * ndim
    coords = np.indices(shape) - radius
    dist = np.sqrt(np.sum(coords**2 / ((radius / 4) ** 2), axis=0))
    res = np.exp(-dist)
    res[res < 0.02] = 0
    return res


def _structure_at_coordinates(
    shape: tuple[int],
    coordinates: np.ndarray,
    structure: np.ndarray,
    *,
    multipliers: Sequence = itertools.repeat(1),
    dtype=None,
    reduce_fn: Callable[
        [np.ndarray, np.ndarray, np.ndarray | None], np.ndarray
    ],
):
    """Update data with structure at given coordinates.

    Parameters
    ----------
    data : ndarray
        Array to update.
    coordinates : ndarray
        Coordinates of the points. The structures will be added at these
        points (center).
    structure : ndarray
        Array with encoded structure. For example, ball (boolean) or density
        (0,1) float.
    multipliers : ndarray
        These values are multiplied by the values in the structure before
        updating the array. Can be used to generate different labels, or to
        vary the intensity of floating point gaussian densities.
    reduce_fn : function
        Function with which to update the array at a particular position. It
        should take two arrays as input and an optional output array.
    """
    radius = (structure.shape[0] - 1) // 2
    data = np.zeros(shape, dtype=dtype)

    for point, value in zip(coordinates, multipliers, strict=False):
        slice_im, slice_ball = _get_slices_at(shape, point, radius)
        reduce_fn(
            data[slice_im], value * structure[slice_ball], out=data[slice_im]
        )
    return data


def _get_slices_at(shape, point, radius):
    slice_im = []
    slice_ball = []
    for i, p in enumerate(point):
        slice_im.append(
            slice(max(0, p - radius), min(shape[i], p + radius + 1))
        )
        ball_start = max(0, radius - p)
        ball_stop = slice_im[-1].stop - slice_im[-1].start + ball_start
        slice_ball.append(slice(ball_start, ball_stop))
    return tuple(slice_im), tuple(slice_ball)


def _update_data_with_mask(data, struct, out=None):
    """Update ``data`` with ``struct`` where ``struct`` is nonzero."""
    # these branches are needed because np.where does not support
    # an out= keyword argument
    if out is None:
        return np.where(struct, struct, data)
    else:  # noqa: RET505
        nz = struct != 0
        out[nz] = struct[nz]
        return out


def _smallest_dtype(n: int) -> np.dtype:
    """Find the smallest dtype that can hold n values."""
    for dtype in [np.uint8, np.uint16, np.uint32, np.uint64]:
        if np.iinfo(dtype).max >= n:
            return dtype
            break
    else:
        raise ValueError(f'{n=} is too large for any dtype.')


@overload
def labeled_particles(
    shape: Sequence[int],
    dtype: np.dtype | None = None,
    n: int = 144,
    seed: int | None = None,
    return_density: Literal[False] = False,
) -> np.ndarray: ...


@overload
def labeled_particles(
    shape: Sequence[int],
    dtype: np.dtype | None = None,
    n: int = 144,
    seed: int | None = None,
    return_density: Literal[True] = True,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]: ...


@lru_cache
def labeled_particles(
    shape: Sequence[int],
    dtype: np.dtype | None = None,
    n: int = 144,
    seed: int | None = None,
    return_density: bool = False,
) -> np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generate labeled blobs of given shape and dtype.

    Parameters
    ----------
    shape : Sequence[int]
        Shape of the resulting array.
    dtype : Optional[np.dtype]
        Dtype of the resulting array.
    n : int
        Number of blobs to generate.
    seed : Optional[int]
        Seed for the random number generator.
    return_density : bool
        Whether to return the density array and center coordinates.
    """
    if dtype is None:
        dtype = _smallest_dtype(n)
    rng = np.random.default_rng(seed)
    ndim = len(shape)
    points = rng.integers(shape, size=(n, ndim))
    values = rng.integers(
        np.iinfo(dtype).min, np.iinfo(dtype).max, size=n, dtype=dtype
    )
    sigma = int(max(shape) / (4.0 * n ** (1 / ndim)))
    ball = _generate_ball(sigma, ndim)

    labels = _structure_at_coordinates(
        shape,
        points,
        ball,
        multipliers=values,
        reduce_fn=_update_data_with_mask,
        dtype=dtype,
    )

    if return_density:
        dens = _generate_density(sigma * 2, ndim)
        densities = _structure_at_coordinates(
            shape, points, dens, reduce_fn=np.maximum, dtype=np.float32
        )

        return labels, densities, points
    else:  # noqa: RET505
        return labels


def run_benchmark_from_module(
    module: ModuleType, klass_name: str, method_name: str
):
    klass = getattr(module, klass_name)
    if getattr(klass, 'params', None):
        skip_if = getattr(klass, 'skip_params', {})
        if isinstance(klass.params[0], Sequence):
            params = itertools.product(*klass.params)
        else:
            params = ((i,) for i in klass.params)
        for param in params:
            if param in skip_if:
                continue
            obj = klass()
            try:
                obj.setup(*param)
            except NotImplementedError:
                continue
            getattr(obj, method_name)(*param)
            getattr(obj, 'teardown', lambda: None)()
    else:
        obj = klass()
        try:
            obj.setup()
        except NotImplementedError:
            return
        getattr(obj, method_name)()
        getattr(obj, 'teardown', lambda: None)()


def run_benchmark(args=None):
    import argparse
    import inspect

    parser = argparse.ArgumentParser(description='Run benchmark')
    parser.add_argument(
        'benchmark', type=str, help='Name of the benchmark to run', default=''
    )

    args = parser.parse_args(args)

    benchmark_selection = args.benchmark.split('.')

    # get module of parent frame
    call_module = inspect.getmodule(inspect.currentframe().f_back)
    run_benchmark_from_module(call_module, *benchmark_selection)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from napari.utils.geometry import (
    point_in_bounding_box,
    project_points_onto_plane,
)

if TYPE_CHECKING:
    from napari.layers.image.image import Image


def displayed_plane_from_nd_line_segment(
    start_point: npt.NDArray,
    end_point: npt.NDArray,
    dims_displayed: list[int] | npt.NDArray,
) -> tuple[npt.NDArray, npt.NDArray]:
    """Get the plane defined by start_point and the normal vector that goes
    from start_point to end_point.

    Note the start_point and end_point are nD and
    the returned plane is in the displayed dimensions (i.e., 3D).

    Parameters
    ----------
    start_point : np.ndarray
        The start point of the line segment in nD coordinates.
    end_point : np.ndarray
        The end point of the line segment in nD coordinates..
    dims_displayed : Union[List[int], np.ndarray]
        The dimensions of the data array currently in view.

    Returns
    -------
    plane_point : np.ndarray
        The point on the plane that intersects the click ray. This is returned
        in data coordinates with only the dimensions that are displayed.
    plane_normal : np.ndarray
        The normal unit vector for the plane. It points in the direction of the click
        in data coordinates.
    """
    plane_point = start_point[dims_displayed]
    end_position_view = end_point[dims_displayed]
    ray_direction = end_position_view - plane_point
    plane_normal = ray_direction / np.linalg.norm(ray_direction)
    return plane_point, plane_normal


def drag_data_to_projected_distance(
    start_position: npt.NDArray,
    end_position: npt.NDArray,
    view_direction: npt.NDArray,
    vector: npt.NDArray,
) -> npt.NDArray:
    """Calculate the projected distance between two mouse events.

    Project the drag vector between two mouse events onto a 3D vector
    specified in data coordinates.

    The general strategy is to
    1) find mouse drag start and end positions, project them onto a
       pseudo-canvas (a plane aligned with the canvas) in data coordinates.
    2) project the mouse drag vector onto the (normalised) vector in data
       coordinates
    Parameters
    ----------
    start_position : np.ndarray
        Starting point of the drag vector in data coordinates
    end_position : np.ndarray
        End point of the drag vector in data coordinates
    view_direction : np.ndarray
        Vector defining the plane normal of the plane onto which the drag
        vector is projected.
    vector : np.ndarray
        (3,) unit vector or (n, 3) array thereof on which to project the drag
        vector from start_event to end_event. This argument is defined in data
        coordinates.
    Returns
    -------
    projected_distance : (1, ) or (n, ) np.ndarray of float
    """
    # enforce at least 2d input
    vector = np.atleast_2d(vector)

    # Store the start and end positions in world coordinates
    start_position = np.asarray(start_position)
    end_position = np.asarray(end_position)

    # Project the start and end positions onto a pseudo-canvas, a plane
    # parallel to the rendered canvas in data coordinates.
    end_position_canvas, _ = project_points_onto_plane(
        end_position, start_position, view_direction
    )
    # Calculate the drag vector on the pseudo-canvas.
    drag_vector_canvas = np.squeeze(end_position_canvas - start_position)

    # Project the drag vector onto the specified vector(s), return the distance
    return np.einsum('j, ij -> i', drag_vector_canvas, vector).squeeze()


def orient_plane_normal_around_cursor(
    layer: Image, plane_normal: tuple
) -> None:
    """Orient a rendering plane by rotating it around the cursor.

    If the cursor ray does not intersect the plane, the position will remain
    unchanged.

    Parameters
    ----------
    layer : Image
        The layer on which the rendering plane is to be rotated
    plane_normal : 3-tuple
        The target plane normal in scene coordinates.
    """
    # avoid circular imports
    import napari
    from napari.layers.image._image_constants import VolumeDepiction

    viewer = napari.viewer.current_viewer()
    if viewer is None:
        return

    # early exit
    if viewer.dims.ndisplay != 3 or layer.depiction != VolumeDepiction.PLANE:
        return

    # find cursor-plane intersection in data coordinates
    cursor_position = layer._world_to_displayed_data(
        position=np.asarray(viewer.cursor.position),
        dims_displayed=layer._slice_input.displayed,
    )
    view_direction = layer._world_to_displayed_data_ray(
        np.asarray(viewer.camera.view_direction), dims_displayed=[-3, -2, -1]
    )
    intersection = layer.plane.intersect_with_line(
        line_position=cursor_position, line_direction=view_direction
    )

    # check if intersection is within data extents for displayed dimensions
    bounding_box = layer.extent.data[:, layer._slice_input.displayed]

    # update plane position
    if point_in_bounding_box(intersection, bounding_box):
        layer.plane.position = intersection

    # update plane normal
    layer.plane.normal = layer._world_to_displayed_data_normal(
        np.asarray(plane_normal), dims_displayed=layer._slice_input.displayed
    )


def nd_line_segment_to_displayed_data_ray(
    start_point: np.ndarray,
    end_point: np.ndarray,
    dims_displayed: list[int] | np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """Convert the start and end point of the line segment of a mouse click ray
    intersecting a data cube to a ray (i.e., start position and direction) in
    displayed data coordinates

    Note: the ray starts 0.1 data units outside of the data volume.

    Parameters
    ----------
    start_point : np.ndarray
        The start position of the ray used to interrogate the data.
    end_point : np.ndarray
        The end position of the ray used to interrogate the data.
    dims_displayed : List[int]
        The indices of the dimensions currently displayed in the Viewer.

    Returns
    -------
    start_position : np.ndarray
        The start position of the ray in displayed data coordinates
    ray_direction : np.ndarray
        The unit vector describing the ray direction.
    """
    # get the ray in the displayed data coordinates
    start_position = start_point[dims_displayed]
    end_position = end_point[dims_displayed]
    ray_direction = end_position - start_position
    ray_direction = ray_direction / np.linalg.norm(ray_direction)
    # step the start position back a little bit to be able to detect shapes
    # that contain the start_position
    start_position = start_position - 0.1 * ray_direction
    return start_position, ray_direction



import numpy as np
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication

from napari._qt.widgets.qt_color_swatch import (
    TRANSPARENT,
    QColorPopup,
    QColorSwatch,
    QColorSwatchEdit,
)


@pytest.mark.parametrize('color', [None, [1, 1, 1, 1]])
@pytest.mark.parametrize('tooltip', [None, 'This is a test'])
def test_succesfull_create_qcolorswatchedit(qtbot, color, tooltip):
    widget = QColorSwatchEdit(initial_color=color, tooltip=tooltip)
    qtbot.add_widget(widget)

    test_color = color or TRANSPARENT
    test_tooltip = tooltip or 'click to set color'

    # check widget creation and base values
    assert widget.color_swatch.toolTip() == test_tooltip
    np.testing.assert_array_equal(widget.color, test_color)

    # check widget popup
    qtbot.mouseRelease(widget.color_swatch, Qt.MouseButton.LeftButton)
    color_popup = None
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QColorPopup):
            color_popup = widget
    assert color_popup


@pytest.mark.parametrize('color', [None, [1, 1, 1, 1]])
@pytest.mark.parametrize('tooltip', [None, 'This is a test'])
def test_succesfull_create_qcolorswatch(qtbot, color, tooltip):
    widget = QColorSwatch(initial_color=color, tooltip=tooltip)
    qtbot.add_widget(widget)

    test_color = color or TRANSPARENT
    test_tooltip = tooltip or 'click to set color'

    # check widget creation and base values
    assert widget.toolTip() == test_tooltip
    np.testing.assert_array_equal(widget.color, test_color)

    # check widget popup
    qtbot.mouseRelease(widget, Qt.MouseButton.LeftButton)
    color_popup = None
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QColorPopup):
            color_popup = widget
    assert color_popup
